import os
import numpy as np
import matplotlib.pyplot as plt

from energy_demand.read_write import read_data, write_data, data_loader
from energy_demand.plotting import basic_plot_functions

print("START")
def plotting_weather_data(path):
    """

    Things to plot

    - annual t_min of all realizations
    - annual t_max of all realizations

    - annual maximum t_max
    - annual minimum t_min
    """
    sim_yrs = range(2015, 2051, 5)
    weather_reationzations = ["NF{}".format(i) for i in range(1, 101,1)]

    container_weather_stations = {}
    container_temp_data = {}

    # All used weather stations from model run
    used_stations = ['station_id_253', 'station_id_252', 'station_id_253', 'station_id_252', 'station_id_252', 'station_id_328', 'station_id_329', 'station_id_305', 'station_id_282', 'station_id_335', 'station_id_335', 'station_id_359', 'station_id_358', 'station_id_309', 'station_id_388', 'station_id_418', 'station_id_420', 'station_id_389', 'station_id_433', 'station_id_385', 'station_id_374', 'station_id_481', 'station_id_481', 'station_id_480', 'station_id_466', 'station_id_531', 'station_id_532', 'station_id_535', 'station_id_535', 'station_id_484', 'station_id_421', 'station_id_472', 'station_id_526', 'station_id_525', 'station_id_526', 'station_id_504', 'station_id_503', 'station_id_504', 'station_id_505', 'station_id_504', 'station_id_504', 'station_id_455', 'station_id_548', 'station_id_546', 'station_id_537', 'station_id_545', 'station_id_236', 'station_id_353', 'station_id_352', 'station_id_384', 'station_id_510', 'station_id_527', 'station_id_550', 'station_id_501', 'station_id_456', 'station_id_472', 'station_id_201', 'station_id_470', 'station_id_487', 'station_id_505', 'station_id_486', 'station_id_457', 'station_id_533', 'station_id_458', 'station_id_441', 'station_id_440', 'station_id_473', 'station_id_217', 'station_id_247', 'station_id_199', 'station_id_232', 'station_id_234', 'station_id_478', 'station_id_248', 'station_id_388', 'station_id_377', 'station_id_376', 'station_id_376', 'station_id_388', 'station_id_354', 'station_id_376', 'station_id_388', 'station_id_515', 'station_id_514', 'station_id_514', 'station_id_531', 'station_id_532', 'station_id_494', 'station_id_512', 'station_id_535', 'station_id_535', 'station_id_517', 'station_id_534', 'station_id_533', 'station_id_549', 'station_id_549', 'station_id_550', 'station_id_549', 'station_id_508', 'station_id_490', 'station_id_507', 'station_id_526', 'station_id_508', 'station_id_491', 'station_id_507', 'station_id_489', 'station_id_509', 'station_id_526', 'station_id_545', 'station_id_492', 'station_id_490', 'station_id_451', 'station_id_467', 'station_id_450', 'station_id_451', 'station_id_466', 'station_id_451', 'station_id_521', 'station_id_538', 'station_id_537', 'station_id_537', 'station_id_522', 'station_id_546', 'station_id_536', 'station_id_522', 'station_id_520', 'station_id_537', 'station_id_488', 'station_id_487', 'station_id_488', 'station_id_472', 'station_id_487', 'station_id_487', 'station_id_551', 'station_id_544', 'station_id_419', 'station_id_525', 'station_id_552', 'station_id_525', 'station_id_543', 'station_id_542', 'station_id_552', 'station_id_543', 'station_id_544', 'station_id_542', 'station_id_542', 'station_id_306', 'station_id_305', 'station_id_282', 'station_id_306', 'station_id_406', 'station_id_264', 'station_id_306', 'station_id_283', 'station_id_284', 'station_id_306', 'station_id_305', 'station_id_304', 'station_id_283', 'station_id_418', 'station_id_403', 'station_id_419', 'station_id_418', 'station_id_403', 'station_id_402', 'station_id_392', 'station_id_379', 'station_id_391', 'station_id_422', 'station_id_404', 'station_id_379', 'station_id_443', 'station_id_444', 'station_id_445', 'station_id_423', 'station_id_469', 'station_id_425', 'station_id_444', 'station_id_461', 'station_id_438', 'station_id_437', 'station_id_439', 'station_id_438', 'station_id_455', 'station_id_454', 'station_id_438', 'station_id_284', 'station_id_268', 'station_id_286', 'station_id_266', 'station_id_288', 'station_id_270', 'station_id_333', 'station_id_389', 'station_id_378', 'station_id_389', 'station_id_389', 'station_id_377', 'station_id_390', 'station_id_403', 'station_id_469', 'station_id_485', 'station_id_484', 'station_id_469', 'station_id_499', 'station_id_498', 'station_id_516', 'station_id_497', 'station_id_479', 'station_id_400', 'station_id_387', 'station_id_401', 'station_id_374', 'station_id_400', 'station_id_386', 'station_id_375', 'station_id_401', 'station_id_491', 'station_id_459', 'station_id_492', 'station_id_476', 'station_id_475', 'station_id_477', 'station_id_462', 'station_id_523', 'station_id_523', 'station_id_522', 'station_id_523', 'station_id_523', 'station_id_523', 'station_id_505', 'station_id_522', 'station_id_541', 'station_id_539', 'station_id_523', 'station_id_417', 'station_id_417', 'station_id_437', 'station_id_436', 'station_id_436', 'station_id_548', 'station_id_547', 'station_id_488', 'station_id_539', 'station_id_540', 'station_id_540', 'station_id_540', 'station_id_547', 'station_id_416', 'station_id_434', 'station_id_435', 'station_id_434', 'station_id_434', 'station_id_415', 'station_id_488', 'station_id_488', 'station_id_489', 'station_id_329', 'station_id_330', 'station_id_330', 'station_id_330', 'station_id_330', 'station_id_329', 'station_id_354', 'station_id_330', 'station_id_329', 'station_id_329', 'station_id_328', 'station_id_328', 'station_id_328', 'station_id_304', 'station_id_327', 'station_id_331', 'station_id_357', 'station_id_356', 'station_id_355', 'station_id_221', 'station_id_221', 'station_id_221', 'station_id_237', 'station_id_416', 'station_id_417', 'station_id_416', 'station_id_416', 'station_id_417', 'station_id_400', 'station_id_400', 'station_id_307', 'station_id_307', 'station_id_331', 'station_id_308', 'station_id_332', 'station_id_221', 'station_id_506', 'station_id_507', 'station_id_506', 'station_id_525', 'station_id_506', 'station_id_524', 'station_id_506', 'station_id_524', 'station_id_505', 'station_id_506', 'station_id_524', 'station_id_506', 'station_id_506', 'station_id_506', 'station_id_505', 'station_id_507', 'station_id_505', 'station_id_505', 'station_id_506', 'station_id_506', 'station_id_523', 'station_id_524', 'station_id_524', 'station_id_524', 'station_id_506', 'station_id_507', 'station_id_505', 'station_id_524', 'station_id_524', 'station_id_506', 'station_id_506', 'station_id_524', 'station_id_506', 'station_id_172', 'station_id_193', 'station_id_173', 'station_id_133', 'station_id_130', 'station_id_168', 'station_id_194', 'station_id_152', 'station_id_170', 'station_id_214', 'station_id_195', 'station_id_103', 'station_id_124', 'station_id_110', 'station_id_176', 'station_id_136', 'station_id_125', 'station_id_121', 'station_id_9', 'station_id_111', 'station_id_105', 'station_id_36', 'station_id_108', 'station_id_52', 'station_id_119', 'station_id_4', 'station_id_94', 'station_id_159', 'station_id_137', 'station_id_102', 'station_id_77', 'station_id_303', 'station_id_2', 'station_id_154', 'station_id_64', 'station_id_89', 'station_id_124', 'station_id_109', 'station_id_109', 'station_id_123', 'station_id_86', 'station_id_105', 'station_id_110', 'station_id_110', 'station_id_448', 'station_id_348', 'station_id_349', 'station_id_350', 'station_id_351', 'station_id_372', 'station_id_394', 'station_id_449', 'station_id_464', 'station_id_407', 'station_id_428', 'station_id_446', 'station_id_446', 'station_id_463', 'station_id_463', 'station_id_464', 'station_id_447', 'station_id_448', 'station_id_448', 'station_id_396', 'station_id_447']

    # Load full data
    for weather_realisation in weather_reationzations:
        print("weather_realisation: " + str(weather_realisation))
        path_weather_data = path

        weather_stations, temp_data = data_loader.load_temp_data(
            {},
            sim_yrs=sim_yrs,
            weather_realisation=weather_realisation,
            path_weather_data=path_weather_data,
            same_base_year_weather=False,
            crit_temp_min_max=True,
            load_np=False,
            load_parquet=False,
            load_csv=True)

        # Load only data from selected weather stations
        temp_data_used = {}
        for year in sim_yrs:
            temp_data_used[year] = {}
            all_station_data = temp_data[year].keys()
            for station in all_station_data:
                if station in used_stations:
                    temp_data_used[year][station] = temp_data[year][station]

        container_weather_stations[weather_realisation] = weather_stations
        container_temp_data[weather_realisation] = temp_data_used

    # Create plot with daily min
    print("... creating min max plot")
    t_min_average_every_day = []
    t_max_average_every_day = []
    t_min_min_every_day = []
    t_max_max_every_day = []
    std_dev_t_min = []
    std_dev_t_max = []
    std_dev_t_min_min = []
    std_dev_t_max_max = []

    for year in sim_yrs:

        for realization in container_weather_stations.keys():
            t_min_average_stations = []
            t_max_average_stations = []
            t_min_min_average_stations = []
            t_max_max_average_stations = []
            stations_data = container_temp_data[realization][year]
            for station in stations_data.keys():
                t_min_annual_average = np.average(stations_data[station]['t_min'])
                t_max_annual_average = np.average(stations_data[station]['t_max'])
                t_min_min_stations = np.min(stations_data[station]['t_min'])
                t_max_max_stations = np.max(stations_data[station]['t_max'])

                t_min_average_stations.append(t_min_annual_average) #average cross all stations
                t_max_average_stations.append(t_max_annual_average) #average cross all stations
                t_min_min_average_stations.append(t_min_min_stations)
                t_max_max_average_stations.append(t_max_max_stations)

        av_t_min = np.average(t_min_average_stations) #average across all realizations
        av_t_max = np.average(t_max_average_stations) #average across all realizations
        av_min_t_min = np.average(t_min_min_average_stations) #average across all realizations
        av_max_t_max = np.average(t_max_max_average_stations) #average across all realizations

        std_t_min = np.std(t_min_average_stations)
        std_t_max = np.std(t_max_average_stations)
        std_t_min_min = np.std(t_min_min_average_stations)
        std_t_max_max = np.std(t_max_max_average_stations)

        t_min_average_every_day.append(av_t_min)
        t_max_average_every_day.append(av_t_max)
        t_min_min_every_day.append(av_min_t_min)
        t_max_max_every_day.append(av_max_t_max)
        
        std_dev_t_min.append(std_t_min)
        std_dev_t_max.append(std_t_max)
        std_dev_t_min_min.append(std_t_min_min)
        std_dev_t_max_max.append(std_t_max_max)

    # Plot variability
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(9, 6)) #width, height

    colors = {
        't_min': 'steelblue',
        't_max': 'tomato',
        't_min_min': 'peru',
        't_max_max': 'r'}

    # plot
    plt.plot(sim_yrs, t_min_average_every_day, color=colors['t_min'], label="t_min")
    plt.plot(sim_yrs, t_max_average_every_day, color=colors['t_max'], label="t_max")
    plt.plot(sim_yrs, t_min_min_every_day, color=colors['t_min_min'], label="t_min_min")
    plt.plot(sim_yrs, t_max_max_every_day, color=colors['t_max_max'], label="t_max_max")

    # Variations
    plt.fill_between(
        sim_yrs,
        list(np.array(t_min_average_every_day) - (2 * np.array(std_dev_t_min))),
        list(np.array(t_min_average_every_day) + (2 * np.array(std_dev_t_min))),
        color=colors['t_min'],
        alpha=0.25)

    plt.fill_between(
        sim_yrs,
        list(np.array(t_max_average_every_day) - (2 * np.array(std_dev_t_max))),
        list(np.array(t_max_average_every_day) + (2 * np.array(std_dev_t_max))),
        color=colors['t_max'],
        alpha=0.25)

    plt.fill_between(
        sim_yrs,
        list(np.array(t_min_min_every_day) - (2 * np.array(std_dev_t_min_min))),
        list(np.array(t_min_min_every_day) + (2 * np.array(std_dev_t_min_min))),
        color=colors['t_min_min'],
        alpha=0.25)

    plt.fill_between(
        sim_yrs,
        list(np.array(t_max_max_every_day) - (2 * np.array(std_dev_t_max_max))),
        list(np.array(t_max_max_every_day) + (2 * np.array(std_dev_t_max_max))),
        color=colors['t_max_max'],
        alpha=0.25)
    
    # Legend
    legend = plt.legend(
        ncol=2,
        prop={'size': 10},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.1),
        frameon=False)
    legend.get_title().set_fontsize(8)

    result_path = "C:/_scrap/"
    seperate_legend = True
    if seperate_legend:
        basic_plot_functions.export_legend(
            legend,
            os.path.join(result_path, "{}__legend.pdf".format(result_path)))
        legend.remove()

    plt.legend(ncol=2)
    plt.xlabel("Year")
    plt.ylabel("Temperature (°C)")
    
    plt.tight_layout()
    plt.margins(x=0)

    fig.savefig(os.path.join(result_path, "test.pdf"))


if __name__ == "__main__":
    """
    """
    plotting_weather_data("//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/J-MARIUS_data/_weather_realisation")


"""Generate scenario paramters for every year
"""
import logging
from collections import defaultdict

from energy_demand.technologies import diffusion_technologies
from energy_demand import enduse_func
from energy_demand.read_write import narrative_related

def get_correct_narrative_timestep(
        sim_yr,
        narrative_timesteps
    ):
    """Based on simulated year select the correct
    narrative timestep, i.e. the year which is
    to be used from the narrative

    Arguments
    ---------
    sim_yr : int
        Simulation year
    narrative_timesteps : list
        All defined timesteps in narrative

    Returns
    -------
    timestep : int
        Narrative timestep to use for calculation

    Example
    -------
    If we have a two-step narrative such as:
        year: 2015 - 2030, 2030 - 2050

    for the sim_yr 2020, the calculated values for
    2030 would need to be used. For the year 2031,
    the values 2050 would need to be used.
    """
    narrative_timesteps.sort()

    # Get corresponding narrative timestep
    if len(narrative_timesteps) == 1:
        timestep = narrative_timesteps[0]

        return timestep
    else:
        # Test if current year is larger than any narrative_timestep
        # and use last defined timestep if this is true. Otherwise
        # get correct timestep from narrative_timesteps
        if sim_yr > narrative_timesteps[-1]:
            timestep = narrative_timesteps[-1]
        else:
            for year_narrative in narrative_timesteps:

                if sim_yr <= year_narrative:
                    timestep = year_narrative

                    return timestep

def calc_annual_switch_params(
        sim_yrs,
        regions,
        diffusion_param_tech,
        base_yr,
        s_tech_by_p,
    ):
    """Calculate annual diffusion parameters
    for technologies based on sigmoid diffusion parameters

    Arguments
    ---------
    sim_yrs : list
        Simulated years
    regions : list
        Regions
    diffusion_param_tech : dict
        Diffusion parameters per submodel

    Returns
    -------
    """
    annual_tech_diff_params = {}

    for region in regions:
        annual_tech_diff_params[region] = {}

        for enduse, sector_region_tech_vals in diffusion_param_tech.items():
            annual_tech_diff_params[region][enduse] = {}

            for sector, reg_vals in sector_region_tech_vals.items():

                if reg_vals != {}:
                    annual_tech_diff_params[region][enduse][sector] = defaultdict(dict)

                    for sim_yr in sim_yrs:

                        #NEW If base year, take base year percentage
                        if sim_yr == base_yr:
                            for tech, base_yr_p in s_tech_by_p[sector][enduse].items():
                                annual_tech_diff_params[region][enduse][sector][tech][sim_yr] = base_yr_p
                        else:
                            narrative_timesteps = list(sector_region_tech_vals[sector].keys())
                            correct_narrative_timestep = get_correct_narrative_timestep(
                                sim_yr=sim_yr, narrative_timesteps=narrative_timesteps)

                            # Calculate parameter for year with diffusion parameter
                            for tech, param in reg_vals[correct_narrative_timestep][region].items():
                                p_s_tech = enduse_func.get_service_diffusion(param, sim_yr)
                                annual_tech_diff_params[region][enduse][sector][tech][sim_yr] = p_s_tech
                else:
                    annual_tech_diff_params[region][enduse][sector] = []

    return annual_tech_diff_params

def generate_annual_param_vals(
        regions,
        strategy_vars,
        sim_yrs,
    ):
    """
    Calculate parameter values for every year based
    on defined narratives and also add a generic
    container of other information necessary
    for parameter.

    Inputs
    -------
    regions : dict
        Regions
    strategy_vars : dict4
        Strategy variable infirmation
    sim_yrs : list
        Simulated years

    Returns
    -------
    container_reg_param : dict
        Values for all simulated years for every region (all parameters for which values
        are provided for every region)
    container_non_reg_param : dict
        Values for all simulated years (all the same for very region)
    """
    reg_param = {}
    for region in regions:
        reg_param[region] = {}
    non_reg_param = {}

    for var_name, strategy_vars_values in strategy_vars.items():

        for region in regions:
            reg_param[region][var_name] = {}
        non_reg_param[var_name] = {}

        single_dim_var = narrative_related.crit_dim_var(
            strategy_vars_values)

        logging.debug("Calculating annual values for parameter: {} {}".format(var_name, single_dim_var))
        if single_dim_var:

            # Additional dictionary passed along with every variable containing additional information
            param_info = {}
            try:
                param_info['fueltype_replace'] = strategy_vars_values[0]['fueltype_replace']
                param_info['fueltype_new'] = strategy_vars_values[0]['fueltype_new']
            except KeyError:
                pass

            # Calculate annual parameter value
            regional_strategy_vary = generate_general_parameter(
                regions=regions,
                narratives=strategy_vars_values,
                sim_yrs=sim_yrs)

            # Test if regional specific or not based on first narrative
            for narrative in strategy_vars_values[:1]:
                reg_specific_crit = narrative['regional_specific']

            if reg_specific_crit:
                for region in regions:
                    reg_param[region][var_name] = regional_strategy_vary[region]
                    reg_param[region][var_name]['param_info'] = param_info
            else:
                non_reg_param[var_name] = regional_strategy_vary
                non_reg_param[var_name]['param_info'] = param_info
        else:
            if not hasattr(strategy_vars_values, 'items'):
                msg = "Error in {}. Object is not a dict, but a {}: {}"
                raise TypeError(msg.format(var_name,
                                           type(strategy_vars_values),
                                           strategy_vars_values))

            for sub_var_name, sector_sub_var_values in strategy_vars_values.items():

                for region in regions:
                    reg_param[region][var_name][sub_var_name] = {}
                non_reg_param[var_name][sub_var_name] = {}

                if type(sector_sub_var_values) is dict:
                    for sector, sub_var_values in sector_sub_var_values.items():

                        if var_name == 'generic_fuel_switch':

                            all_new_fueltypes = []
                            for narrative in sub_var_values:
                                all_new_fueltypes.append(narrative['fueltype_new'])

                            for new_fueltype in all_new_fueltypes:

                                # Get all narratives of this fueltype
                                narratives = []
                                for narrative in sub_var_values:
                                    if narrative['fueltype_new'] == new_fueltype:
                                        narratives.append(narrative)

                                param_info = {}
                                param_info['fueltype_new'] = int(new_fueltype)
                                param_info['fueltype_replace'] = narrative['fueltype_replace']

                                # Calculate annual parameter value
                                regional_strategy_vary = generate_general_parameter(
                                    regions=regions,
                                    narratives=narratives,
                                    sim_yrs=sim_yrs)

                                if narrative['regional_specific']:
                                    for region in regions:
                                        reg_param[region][var_name][sub_var_name][sector][param_info['fueltype_new']] = dict(regional_strategy_vary[region])
                                        reg_param[region][var_name][sub_var_name][sector][param_info['fueltype_new']]['param_info'] = param_info
                                else:
                                    non_reg_param[var_name][sub_var_name][sector][param_info['fueltype_new']] = regional_strategy_vary
                                    non_reg_param[var_name][sub_var_name][sector][param_info['fueltype_new']]['param_info'] = param_info
                        else:
                            # Calculate annual parameter value
                            regional_strategy_vary = generate_general_parameter(
                                regions=regions,
                                narratives=sub_var_values,
                                sim_yrs=sim_yrs)

                            # Test if regional specific or not based on first narrative
                            for narrative in sub_var_values[:1]:
                                reg_specific_crit = narrative['regional_specific']

                            if reg_specific_crit:
                                for region in regions:
                                    reg_param[region][var_name][sub_var_name][sector] = dict(regional_strategy_vary[region])
                            else:
                                non_reg_param[var_name][sub_var_name][sector] = regional_strategy_vary
                # Single dimensional
                else:
                    param_info = {}
                    if var_name == 'generic_fuel_switch':

                        all_new_fueltypes = []
                        for narrative in sector_sub_var_values:
                            all_new_fueltypes.append(narrative['fueltype_new'])

                        for new_fueltype in all_new_fueltypes:
                            narratives = []
                            for narrative in sector_sub_var_values:
                                if narrative['fueltype_new'] == new_fueltype:
                                    narratives.append(narrative)

                            # Calculate annual parameter value
                            regional_strategy_vary = generate_general_parameter(
                                regions=regions,
                                narratives=narratives,
                                sim_yrs=sim_yrs)

                            param_info = {}
                            param_info['fueltype_replace'] = narrative['fueltype_replace']
                            param_info['fueltype_new'] = new_fueltype

                            if narrative['regional_specific']:
                                for region in regions:
                                    reg_param[region][var_name][sub_var_name][param_info['fueltype_new']] = dict(regional_strategy_vary[region])
                                    reg_param[region][var_name][sub_var_name][param_info['fueltype_new']]['param_info'] = param_info
                            else:
                                non_reg_param[var_name][sub_var_name][param_info['fueltype_new']] = regional_strategy_vary
                                non_reg_param[var_name][sub_var_name][param_info['fueltype_new']]['param_info'][param_info['fueltype_replace']] = param_info
                    else:
                        # Calculate annual parameter value
                        regional_strategy_vary = generate_general_parameter(
                            regions=regions,
                            narratives=sector_sub_var_values,
                            sim_yrs=sim_yrs)

                        # Test if regional specific or not based on first narrative
                        for narrative in sector_sub_var_values[:1]:
                            reg_specific_crit = narrative['regional_specific']

                        if reg_specific_crit:
                            for region in regions:
                                reg_param[region][var_name][sub_var_name] = dict(regional_strategy_vary[region])
                        else:
                            non_reg_param[var_name][sub_var_name] = regional_strategy_vary

    return dict(reg_param), dict(non_reg_param)

def generate_general_parameter(
        regions,
        narratives,
        sim_yrs
    ):
    """Based on narrative input, calculate
    the parameter value for every modelled year

    Arguments
    ---------
    regions : list
        Regions
    narratives : List
        List containing all narratives of how a model
        parameter changes over time
    sim_yrs : list
        Simulated years

    Returns
    --------
    container : dict
        All model paramters containing either:
        - all values for each region and year
        - all values for each region
    """
    container = defaultdict(dict)

    # Get latest narrative timestep as it could be that the narrative is
    # not defined as long enough as the simulated years (e.g. narrative only up
    # to 2040 but the year 2050 is simulated). In these cases, use the largest
    # narrative timestep (maximum assumed to stay constant from this time onwards)
    latest_narrative_timestep = 0
    for narrative in narratives:
        if narrative['end_yr'] > latest_narrative_timestep:
            latest_narrative_timestep = narrative['end_yr']

    for sim_yr in sim_yrs:

        # Set curry_yr to largest year defined narrative if sim_yr is larger
        if sim_yr > latest_narrative_timestep:
            curr_yr = latest_narrative_timestep
        else:
            curr_yr = sim_yr

        for narrative in narratives:
            narrative_yrs = range(narrative['base_yr'], narrative['end_yr'] + 1, 1) # Years which narrative covers

            if curr_yr in narrative_yrs:
                if not narrative['regional_specific']:
                    if narrative['diffusion_choice'] == 'linear':
                        change_cy = diffusion_technologies.linear_diff(
                            narrative['base_yr'],
                            curr_yr,
                            narrative['regional_vals_by'],
                            narrative['regional_vals_ey'],
                            narrative['end_yr'])

                    elif narrative['diffusion_choice'] == 'sigmoid':

                        diff_value = narrative['regional_vals_ey'] - narrative['regional_vals_by']
                        sig_diff_factor = diffusion_technologies.sigmoid_diffusion(
                            narrative['base_yr'],
                            curr_yr,
                            narrative['end_yr'],
                            narrative['sig_midpoint'],
                            narrative['sig_steepness'])
                        change_cy = narrative['regional_vals_by'] + (diff_value * sig_diff_factor)

                    container[sim_yr] = change_cy
                else:
                    for region in regions:

                        if narrative['diffusion_choice'] == 'linear':
                            change_cy = diffusion_technologies.linear_diff(
                                narrative['base_yr'],
                                curr_yr,
                                narrative['regional_vals_by'][region],
                                narrative['regional_vals_ey'][region],
                                narrative['end_yr'])
                        elif narrative['diffusion_choice'] == 'sigmoid':
                            diff_value = narrative['regional_vals_ey'][region] - narrative['regional_vals_by'][region]
                            sig_diff_factor = diffusion_technologies.sigmoid_diffusion(
                                narrative['base_yr'],
                                curr_yr,
                                narrative['end_yr'],
                                narrative['sig_midpoint'],
                                narrative['sig_steepness'])
                            change_cy = narrative['regional_vals_by'][region] + (diff_value * sig_diff_factor)

                        container[region][sim_yr] = change_cy

    return dict(container)




"""
"""
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import copy
import pylab
import numpy as np

from energy_demand.plotting import fig_3_weather_map
from energy_demand.plotting import basic_plot_functions
from energy_demand.basic import conversions
from energy_demand.plotting import plotting_styles
from energy_demand.read_write import write_data

colors = {
    # High elec
    'h_min': 'darkgreen', ##2c7bb6', ##004529',
    'h_c': '#238443',
    'h_max': '#d7191c', #'#78c679',

    # Low elec
    'l_min': 'steelblue', #abd9e9', ##800026',
    'l_c': '#e31a1c',
    'l_max': 'darkorchid', #fdae61', ##fd8d3c',

    'other1': '#C044FF',
    'other2': '#3DF735',
    'other3': '#AD6D70',
    'other4': '#EC2504',
    'other5': '#8C0B90'}

colors_quadrates = {
    0: '#f7f7f7', #Threshold change limit
    1: '#e66101', #'red',
    2: '#fdb863', #'tomato',
    3: '#5e3c99', #'seagreen',
    4: '#b2abd2'} #'orange'}

marker_list = plotting_styles.marker_list()

marker_styles = {
    'h_max': marker_list[0],
    'h_min': marker_list[1],
    'l_min': marker_list[2],
    'l_max': marker_list[6],
}

def clasify_color(diff_mean, diff_std, threshold):
    if diff_mean < -1 * threshold and diff_std < -1 * threshold:
        color_pos = 3
    elif diff_mean > threshold and diff_std < -1 * threshold:
        color_pos = 2
    elif diff_mean < -1 * threshold and diff_std > threshold:
        color_pos = 4
    elif diff_mean > threshold and diff_std > threshold:
        color_pos = 1
    else:
        color_pos = 0

    '''if diff_mean < 0 and diff_std < 0:
        color_pos = 3
    elif diff_mean > 0 and diff_std < 0:
        color_pos = 2
    elif diff_mean < 0 and diff_std > 0:
        color_pos = 4
    elif diff_mean > 0 and diff_std > 0:
        color_pos = 1
        raise Exception("Something strange happend: {} {}".format(diff_mean, diff_std))'''

    return color_pos 

def scenario_over_time(
        scenario_result_container,
        field_name,
        sim_yrs,
        fig_name,
        result_path,
        plot_points,
        crit_smooth_line=True,
        seperate_legend=False
    ):
    """Plot peak over time
    """
    statistics_to_print = []

    fig = plt.figure(figsize=basic_plot_functions.cm2inch(10, 10)) #width, height

    ax = fig.add_subplot(1, 1, 1)

    for cnt_scenario, i in enumerate(scenario_result_container):
        scenario_name = i['scenario_name']
        national_peak = i[field_name]

        # dataframe with national peak (columns= simulation year, row: Realisation) 
        # Calculate quantiles
        #quantile_95 = 0.68 #0.95 #0.68
        #quantile_05 = 0.32 #0.05 #0.32

        try:
            color = colors[scenario_name]
            marker = marker_styles[scenario_name]
        except KeyError:
            color = list(colors.values())[cnt_scenario]

        try:
            marker = marker_styles[scenario_name]
        except KeyError:
            marker = list(marker_styles.values())[cnt_scenario]

        #print("SCENARIO NAME {}  {}".format(scenario_name, color))

        # Calculate average across all weather scenarios
        mean_national_peak = national_peak.mean(axis=0)

        mean_national_peak_sim_yrs = copy.copy(mean_national_peak)

        statistics_to_print.append("scenario: {} values over years: {}".format(scenario_name, mean_national_peak_sim_yrs))

        # Standard deviation over all realisations
        #df_q_05 = national_peak.quantile(quantile_05)
        #df_q_95 = national_peak.quantile(quantile_95)

        # Number of sigma
        nr_of_sigma = 1
        std_dev = national_peak.std(axis=0)
        two_std_line_pos = mean_national_peak + (nr_of_sigma * std_dev)
        two_std_line_neg = mean_national_peak - (nr_of_sigma * std_dev)

        # Maximum and minium values
        max_values = np.max(national_peak)
        min_values = np.min(national_peak)

        statistics_to_print.append("scenario: {} two_sigma_pos: {}".format(scenario_name, two_std_line_pos))
        statistics_to_print.append("scenario: {} two_sigma_neg: {}".format(scenario_name, two_std_line_neg))
        statistics_to_print.append("--------min-------------- {}".format(scenario_name))
        statistics_to_print.append("{}".format(np.min(national_peak))) #Get minimum value for every simulation year of all realizations
        statistics_to_print.append("--------max-------------- {}".format(scenario_name))
        statistics_to_print.append("{}".format(np.max(national_peak)))

        # --------------------
        # Try to smooth lines
        # --------------------
        sim_yrs_smoothed = sim_yrs
        if crit_smooth_line:
            try:
                sim_yrs_smoothed, mean_national_peak_smoothed = basic_plot_functions.smooth_data(sim_yrs, mean_national_peak, num=40000)
                #_, df_q_05_smoothed = basic_plot_functions.smooth_data(sim_yrs, df_q_05, num=40000)
                #_, df_q_95_smoothed = basic_plot_functions.smooth_data(sim_yrs, df_q_95, num=40000)
                _, two_std_line_pos_smoothed = basic_plot_functions.smooth_data(sim_yrs, two_std_line_pos, num=40000)
                _, two_std_line_neg_smoothed = basic_plot_functions.smooth_data(sim_yrs, two_std_line_neg, num=40000)
                _, max_values_smoothed = basic_plot_functions.smooth_data(sim_yrs, max_values, num=40000)
                _, min_values_smoothed = basic_plot_functions.smooth_data(sim_yrs, min_values, num=40000)
                mean_national_peak = pd.Series(mean_national_peak_smoothed, sim_yrs_smoothed)
                #df_q_05 = pd.Series(df_q_05_smoothed, sim_yrs_smoothed)
                #df_q_95 = pd.Series(df_q_95_smoothed, sim_yrs_smoothed)
                two_std_line_pos = pd.Series(two_std_line_pos_smoothed, sim_yrs_smoothed)
                two_std_line_neg = pd.Series(two_std_line_neg_smoothed, sim_yrs_smoothed)

                max_values = pd.Series(max_values_smoothed, sim_yrs_smoothed).values
                min_values = pd.Series(min_values_smoothed, sim_yrs_smoothed).values
            except:
                sim_yrs_smoothed = sim_yrs

        # -----------------------
        # Plot lines
        # ------------------------
        plt.plot(
            mean_national_peak,
            label="{} (mean)".format(scenario_name),
            zorder=1,
            color=color)

        # ------------------------
        # Plot markers
        # ------------------------
        if plot_points:
            plt.scatter(
                sim_yrs,
                mean_national_peak_sim_yrs,
                c=color,
                marker=marker,
                edgecolor='black',
                linewidth=0.5,
                zorder=2,
                s=15,
                clip_on=False) #do not clip points on axis

        # ------------------
        # Start with uncertainty one model step later (=> 2020)
        # ------------------
        start_yr_uncertainty = 2020

        if crit_smooth_line:
            #Get position in array of start year uncertainty
            pos_unc_yr = len(np.where(sim_yrs_smoothed < start_yr_uncertainty)[0])
        else:
            pos_unc_yr = 0
            for cnt, year in enumerate(sim_yrs_smoothed):
                if year == start_yr_uncertainty:
                    pos_unc_yr = cnt

        # select based on index which is year
        #df_q_05 = df_q_05.loc[start_yr_uncertainty:]
        #df_q_95 = df_q_95.loc[start_yr_uncertainty:]
        two_std_line_pos = two_std_line_pos.loc[start_yr_uncertainty:]
        two_std_line_neg = two_std_line_neg.loc[start_yr_uncertainty:]
        sim_yrs_smoothed = sim_yrs_smoothed[pos_unc_yr:]

        min_values = min_values[pos_unc_yr:] #min_values.loc[start_yr_uncertainty:]
        max_values = max_values[pos_unc_yr:] #max_values.loc[start_yr_uncertainty:]

        # --------------------------------------
        # Plottin qunatilse and average scenario
        # --------------------------------------
        #df_q_05.plot.line(color=color, linestyle='--', linewidth=0.1, label='_nolegend_')
        #df_q_95.plot.line(color=color, linestyle='--', linewidth=0.1, label='_nolegend_')

        # Plot standard deviation
        #two_std_line_pos.plot.line(color=color, linestyle='--', linewidth=0.1, label='_nolegend_')
        #two_std_line_neg.plot.line(color=color, linestyle='--', linewidth=0.1, label='_nolegend_')

        # plot min and maximum values
        plt.plot(sim_yrs_smoothed, min_values, color=color, linestyle='--', linewidth=0.3, label='_nolegend_')
        plt.plot(sim_yrs_smoothed, max_values, color=color, linestyle='--', linewidth=0.3, label='_nolegend_')

        plt.fill_between(
            sim_yrs_smoothed,
            list(two_std_line_pos),
            list(two_std_line_neg),
            alpha=0.25,
            facecolor=color)

    plt.xlim(2015, 2050)
    plt.ylim(0)

    # --------
    # Different style
    # --------
    ax = plt.gca()
    ax.grid(which='major', color='black', axis='y', linestyle='--')

    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['top'].set_visible(False)

    plt.tick_params(
        axis='y',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom=False,      # ticks along the bottom edge are off
        top=False,         # ticks along the top edge are off
        left=False,
        right=False,
        labelbottom=False,
        labeltop=False,
        labelleft=True,
        labelright=False) # labels along the bottom edge are off

    # --------
    # Legend
    # --------
    legend = plt.legend(
        ncol=2,
        prop={'size': 10},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.1),
        frameon=False)
    legend.get_title().set_fontsize(8)

    if seperate_legend:
        basic_plot_functions.export_legend(
            legend,
            os.path.join(result_path, "{}__legend.pdf".format(fig_name[:-4])))
        legend.remove()

    # --------
    # Labeling
    # --------
    plt.ylabel("national peak demand (GW)")
    plt.tight_layout()
    plt.savefig(os.path.join(result_path, fig_name))
    plt.close()

    # Write info to txt
    write_data.write_list_to_txt(
        os.path.join(result_path, fig_name).replace(".pdf", ".txt"),
        statistics_to_print)

def plot_std_dev_vs_contribution(
        scenario_result_container,
        sim_yrs,
        fig_name,
        fueltypes,
        result_path,
        plot_points=False,
        path_shapefile_input=False,
        unit='TWh',
        crit_smooth_line=True,
        seperate_legend=False):
    
    colors = {
        2020: 'dimgrey',
        2050: 'forestgreen'
    }

    #plot_numbers = {
    #    'h_max': {'row': 0, 'col': 0},
    #    'h_min': {'row': 0, 'col': 1},
    #    'l_max': {'row': 1, 'col': 0},
    #    'l_min': {'row': 1, 'col': 1}}
    plot_numbers = {
        'h_max': 1,
        'h_min': 2,
        'l_max': 3,
        'l_min': 4}

    sizes = {
        2020: 2,
        2050: 5}

    scenarios = ['l_max', 'l_min', 'h_max', 'h_min']

    threshold = 5 # percent

    # -----------
    # Plot 4x4 chart with relative differences
    # ------------
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(20, 20)) #width, height
    for scenario in scenarios:
        plot_nr = plot_numbers[scenario]
        plt.subplot(2, 2, plot_nr)

        mean_2020 = scenario_result_container[2020][scenario]['regional_share_national_peak'].mean(axis=0)
        mean_2050 = scenario_result_container[2050][scenario]['regional_share_national_peak'].mean(axis=0)
        
        std_2020 = scenario_result_container[2020][scenario]['regional_share_national_peak'].std(axis=0)
        std_2050 = scenario_result_container[2050][scenario]['regional_share_national_peak'].std(axis=0)
        
        diff_2020_2050_reg_share = ((100 / mean_2020) * mean_2050) - 100
        diff_2020_2050_reg_share_std = ((100 / std_2020) * std_2050) - 100
        
        if scenario not in ['h_min', 'l_min']:
            plt.ylabel("Δ standard deviation [%]", fontsize=10)
        if scenario not in ['h_max', 'h_min']:
            plt.xlabel("Δ mean of total peak [%]", fontsize=10)

        regions = diff_2020_2050_reg_share.index
        for region in regions.values:
            diff_mean = diff_2020_2050_reg_share.loc[region]
            diff_std = diff_2020_2050_reg_share_std.loc[region]

            color_pos = clasify_color(diff_mean, diff_std, threshold=threshold)

            color = colors_quadrates[color_pos]
            plt.scatter(
                diff_mean,
                diff_std,
                #alpha=0.6,
                color=color,
                edgecolor=color,
                linewidth=0.5,
                s=3)

        plt.xlim(-30, 60)
        plt.ylim(-50, 300)
        plt.title("{}".format(scenario), size=10)
        plt.axhline(y=0, linewidth=0.7, color='grey', linestyle='--')   # horizontal line
        plt.axvline(x=0, linewidth=0.7, color='grey', linestyle='--')    # vertical line

        # -----------------
        # Plot spatial map
        # -----------------
        regions = diff_2020_2050_reg_share.index
        relclassified_values = pd.DataFrame()
        relclassified_values['reclassified'] = 0
        relclassified_values['name'] = regions
        relclassified_values = relclassified_values.set_index(('name'))
        relclassified_values['name'] = regions

        for region in regions.values:
            diff_mean = diff_2020_2050_reg_share.loc[region]
            diff_std = diff_2020_2050_reg_share_std.loc[region]
            relclassified_values.loc[region, 'reclassified'] = clasify_color(diff_mean, diff_std, threshold=threshold)

        fig_3_weather_map.plot_4_cross_map(
            cmap_rgb_colors=colors_quadrates,
            reclassified=relclassified_values,
            result_path=os.path.join(result_path, "spatial_final_{}.pdf".format(scenario)),
            threshold=threshold,
            path_shapefile_input=path_shapefile_input,
            seperate_legend=True)

    plt.savefig(os.path.join(result_path, "cross_chart_relative.pdf"))

    # -----------
    # Plot 4x4 chart with absolute
    # ------------
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(20, 20)) #width, height
    for year, scenario_results in scenario_result_container.items():
        for scenario_name, scenario_results in scenario_results.items():

            plot_nr = plot_numbers[scenario_name]
            plt.subplot(2, 2, plot_nr)

            if year in [2020, 2050]:
                #regions = list(regional_share_national_peak.columns)
                #nr_of_regions = regional_share_national_peak.shape[1]
                #nr_of_realisations = regional_share_national_peak.shape[0]

                # Mean over all realisations
                mean = scenario_results['regional_share_national_peak'].mean(axis=0) / 100

                # Standard deviation over all realisations
                std_dev = scenario_results['regional_share_national_peak'].std(axis=0) / 100

                # Convert standard deviation given as % of peak into GW (multiply national peak per region share across the columns)
                abs_gw = pd.DataFrame()
                national_peak_per_run = scenario_results['peak_hour_demand'].sum(axis=1).to_frame()

                for reg_column in scenario_results['regional_share_national_peak'].columns.values:

                    #reg share * national peak
                    column_national_peak = national_peak_per_run.columns[0]
                    abs_gw[reg_column] = (scenario_results['regional_share_national_peak'][reg_column] / 100) * national_peak_per_run[column_national_peak]

                # Absolute values
                std_dev_gw = abs_gw.std(axis=0)
                mean_gw = abs_gw.mean(axis=0)
                plt.ylim(0, np.max(std_dev_gw))
                plt.ylim(0, 0.06)
                plt.xlim(0, 1.5)

                if scenario_name not in ['h_min', 'l_min']:
                    plt.ylabel("standard deviation [GW]", fontsize=10)
                if scenario_name not in ['h_max', 'h_min']:
                    plt.xlabel("mean of total peak [GW]", fontsize=10)

                plt.title(scenario_name, size=10)
                plt.scatter(
                    list(mean_gw), #list(mean),
                    list(std_dev_gw), #list(std_dev),
                    color=colors[year],
                    s=sizes[year]) #,
                    #label="{} {}".format(year, scenario_name))

    plt.savefig(os.path.join(result_path, "cross_chart_absolute.pdf"))

def fueltypes_over_time(
        scenario_result_container,
        sim_yrs,
        fig_name,
        fueltypes,
        result_path,
        plot_points=False,
        unit='TWh',
        crit_smooth_line=True,
        seperate_legend=False
    ):
    """Plot fueltypes over time
    """
    statistics_to_print = []

    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(10, 10)) #width, height
    ax = fig.add_subplot(1, 1, 1)

    colors = {
        # Low elec
        'electricity':  '#3e3838',
        'gas':          '#ae7c7c',
        'hydrogen':     '#6cbbb3',
    }

    line_styles_default = plotting_styles.linestyles()

    linestyles = {
        'h_max': line_styles_default[0],
        'h_min': line_styles_default[6],
        'l_min': line_styles_default[8],
        'l_max': line_styles_default[9],
    }

    for cnt_scenario, i in enumerate(scenario_result_container):
        scenario_name = i['scenario_name']

        for cnt_linestyle, fueltype_str in enumerate(fueltypes):
            national_sum = i['national_{}'.format(fueltype_str)]

            if unit == 'TWh':
                unit_factor = conversions.gwh_to_twh(1)
            elif unit == 'GWh':
                unit_factor = 1
            else:
                raise Exception("Wrong unit")

            national_sum = national_sum * unit_factor

            try:
                color = colors[fueltype_str]
            except KeyError:
                raise Exception("Wrong color")

            try:
                linestyle = linestyles[scenario_name]
            except KeyError:
                linestyle = list(linestyles.values())[cnt_scenario]
            try:
                marker = marker_styles[scenario_name]
            except KeyError:
                marker = list(marker_styles.values())[cnt_scenario]

            # Calculate average across all weather scenarios
            mean_national_sum = national_sum.mean(axis=0)
            mean_national_sum_sim_yrs = copy.copy(mean_national_sum)

            statistics_to_print.append("{} fueltype_str: {} mean_national_sum_sim_yrs: {}".format(scenario_name, fueltype_str, mean_national_sum_sim_yrs))

            # Standard deviation over all realisations
            std = np.std(national_sum)

            nr_of_sigma = 1
            pos_sigma = mean_national_sum + (nr_of_sigma * std)
            neg_sigma = mean_national_sum - (nr_of_sigma * std)

            statistics_to_print.append("{} fueltype_str: {} pos_sigma: {}".format(scenario_name, fueltype_str, pos_sigma))
            statistics_to_print.append("{} fueltype_str: {} neg_sigma: {}".format(scenario_name, fueltype_str, neg_sigma))

            # --------------------
            # Try to smooth lines
            # --------------------
            sim_yrs_smoothed = sim_yrs
            if crit_smooth_line:
                try:
                    sim_yrs_smoothed, mean_national_sum_smoothed = basic_plot_functions.smooth_data(sim_yrs, mean_national_sum, num=500)
                    _, pos_sigma_smoothed = basic_plot_functions.smooth_data(sim_yrs, pos_sigma, num=500)
                    _, neg_sigma_smoothed = basic_plot_functions.smooth_data(sim_yrs, neg_sigma, num=500)
                    mean_national_sum = pd.Series(mean_national_sum_smoothed, sim_yrs_smoothed)                 
                    
                    pos_sigma = pd.Series(pos_sigma_smoothed, sim_yrs_smoothed).values
                    neg_sigma = pd.Series(neg_sigma_smoothed, sim_yrs_smoothed).values
                except:
                    pos_sigma = pos_sigma.values
                    neg_sigma = neg_sigma.values
            
            # ------------------------
            # Plot lines
            # ------------------------
            plt.plot(
                mean_national_sum,
                label="{} {}".format(fueltype_str, scenario_name),
                linestyle=linestyle,
                color=color,
                zorder=1,
                clip_on=True)

            # ------------------------
            # Plot markers
            # ------------------------
            if plot_points:
                plt.scatter(
                    sim_yrs,
                    mean_national_sum_sim_yrs,
                    marker=marker,
                    edgecolor='black',
                    linewidth=0.5,
                    c=color,
                    zorder=2,
                    s=15,
                    clip_on=False) #do not clip points on axis

            # ------------------
            # Start with uncertainty one model step later (=> 2020)
            # ------------------
            start_yr_uncertainty = 2020

            if crit_smooth_line:
                #Get position in array of start year uncertainty
                pos_unc_yr = len(np.where(sim_yrs_smoothed < start_yr_uncertainty)[0])
            else:
                pos_unc_yr = 0
                for cnt, year in enumerate(sim_yrs_smoothed):
                    if year == start_yr_uncertainty:
                        pos_unc_yr = cnt

            # Shorten lines
            #pos_sigma = pos_sigma.loc[start_yr_uncertainty:]
            #neg_sigma = neg_sigma.loc[start_yr_uncertainty:]
            pos_sigma = pos_sigma[pos_unc_yr:]
            neg_sigma = neg_sigma[pos_unc_yr:]

            sim_yrs_smoothed = sim_yrs_smoothed[pos_unc_yr:]

            plt.plot(sim_yrs_smoothed, pos_sigma, color=color, linestyle='--', linewidth=0.1, label='_nolegend_')
            plt.plot(sim_yrs_smoothed, neg_sigma, color=color, linestyle='--', linewidth=0.1, label='_nolegend_')

            # Plotting qunatilse and average scenario
            plt.fill_between(
                sim_yrs_smoothed,
                list(pos_sigma),  #y1
                list(neg_sigma),  #y2
                alpha=0.25,
                facecolor=color)

    plt.xlim(2015, 2050)
    plt.ylim(0)

    ax = plt.gca()

    # Major ticks every 20, minor ticks every 5
    major_ticks = [200, 400, 600] #np.arange(0, 600, 200)
    minor_ticks = [100, 200, 300, 400, 500, 600] #np.arange(0, 600, 100)

    ax.set_yticks(major_ticks)
    ax.set_yticks(minor_ticks, minor=True)

    # And a corresponding grid
    ax.grid(
        which='both',
        color='black',
        linewidth=0.5,
        axis='y',
        linestyle=line_styles_default[3]) #[6])

    # Or if you want different settings for the grids:
    ax.grid(which='minor', axis='y', alpha=0.4)
    ax.grid(which='major', axis='y', alpha=0.8)

    # Achsen
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['top'].set_visible(False)

    # Ticks
    plt.tick_params(
        axis='y',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom=False,      # ticks along the bottom edge are off
        top=False,         # ticks along the top edge are off
        left=False,
        right=False,
        labelbottom=False,
        labeltop=False,
        labelleft=True,
        labelright=False) # labels along the bottom edge are off

    # --------
    # Legend
    # --------
    legend = plt.legend(
        ncol=2,
        prop={'size': 6},
        loc='upper center',
        bbox_to_anchor=(0.5, -0.1),
        frameon=False)
    legend.get_title().set_fontsize(8)

    if seperate_legend:
        basic_plot_functions.export_legend(
            legend,
            os.path.join(result_path, "{}__legend.pdf".format(fig_name[:-4])))
        legend.remove()

    # --------
    # Labeling
    # --------
    plt.ylabel("national fuel over time [in {}]".format(unit))
    #plt.xlabel("year")
    #plt.title("Title")

    plt.tight_layout()
    #plt.show()
    plt.savefig(os.path.join(result_path, fig_name))
    plt.close()

    write_data.write_list_to_txt(
        os.path.join(result_path, fig_name).replace(".pdf", ".txt"),
        statistics_to_print)



"""Testing functions ``dwelling_stock`` ``dw_stock``
"""
from energy_demand.dwelling_stock import dw_stock

def test_get_tot_pop():
    """Testing
    """
    scenario_drivers = {'heating': ['population']}
    classobject1 = dw_stock.Dwelling(
        2015,
        {'longitude': 10, 'latitude': 10},
        1000,
        ['heating'],
        scenario_drivers,
        population=2.2
    )

    classobject2 = dw_stock.Dwelling(
        2015,
        {'longitude': 10, 'latitude': 10},
        1000,
        ['heating'],
        scenario_drivers
    )
    dwellings = [classobject1, classobject1]
    dwellings2 = [classobject2, classobject2]

    dw_stock_object = dw_stock.DwellingStock(dwellings, ['heating'])
    dw_stock_object2 = dw_stock.DwellingStock(dwellings2, ['heating'])
    
    expected = 4.4
    expected2 = None

    # call function
    out_value = dw_stock.get_tot_pop(dw_stock_object.dwellings) 
    out_value2 = dw_stock.get_tot_pop(dw_stock_object2.dwellings)

    assert out_value == expected
    assert out_value2 == expected2

def test_get_scenario_driver():
    """Testing
    """
    scenario_drivers = {'heating': ['population']}
    classobject1 = dw_stock.Dwelling(
        2015,
        {'longitude': 10, 'latitude': 10},
        1000,
        ['heating'],
        scenario_drivers,
        population=2.2)

    dwellings = [classobject1, classobject1]
    dw_stock_object = dw_stock.DwellingStock(dwellings, ['heating'])

    expected = 4.4

    # call function
    out_value = dw_stock_object.get_scenario_driver('heating')
    #out_value = dw_stock.get_scenario_driver(dw_stock_object.dwellings, 'population')

    assert out_value == expected

def test_dwelling():
    """Testing
    """
    scenario_drivers = {'heating': ['population']}
    classobject = dw_stock.Dwelling(
        2015,
        {'longitude': 10, 'latitude': 10},
        1000,
        ['heating'],
        scenario_drivers,
        population=2.2
    )

    # Driver
    expected = 2.2

    # call function
    classobject.calc_scenario_driver(scenario_drivers)

    out_value = classobject.heating

    assert out_value == expected



"""Script to disaggregate national data into regional data
"""
import logging
import copy
from collections import defaultdict
import numpy as np

from energy_demand.profiles import hdd_cdd
from energy_demand.basic import testing_functions
from energy_demand.basic import lookup_tables
from energy_demand.read_write import data_loader
from energy_demand.scripts import init_scripts
from energy_demand.technologies import tech_related

def disaggr_demand(data, crit_temp_min_max, spatial_calibration=False):
    """Disaggregated demand

    Arguments
    ----------
    data : dict
        Data

    Returns
    --------
    disagg : dict
        Disaggregated energy demand
    spatial_calibration : bool
        Criteria wheter base year data should be used to
        calibrate spatial disaggregation (non-residential demand)
    """
    disagg = {}

    # ===========================================
    # I. Disaggregation
    # ===========================================
    base_yr = data['assumptions'].base_yr

    # Load data for disaggregateion
    data['scenario_data']['employment_stats'] = data_loader.read_employment_stats(
        data['paths']['path_employment_statistics'])

    # Disaggregate fuel for all regions
    disagg['residential'], disagg['service'], disagg['industry'] = disaggregate_base_demand(
        data['pop_for_disag'],
        data['regions'],
        data['fuels'],
        data['scenario_data'],
        data['assumptions'],
        data['reg_coord'],
        data['weather_stations'], # Base year data used to disaggregate demand
        data['temp_data'][base_yr], # Base year data used to disaggregate demand
        data['sectors'],
        data['enduses'],
        data['service_building_count'],
        crit_temp_min_max)

    if spatial_calibration:
        '''The spatial disaggregation of non-residential demand
        can be calibrated for gas and electricity based on actual
        measured demand data

        Note: All other fueltypes are not scaled
        '''
        calibrate_residential = False # Calibrate residential demands
        calibrate_non_residential = True # Calibrate non residential demands

        # Non-residential electricity regional demands of base year for electrictiy and gas
        fueltype_elec = tech_related.get_fueltype_int('electricity')
        fueltype_gas = tech_related.get_fueltype_int('gas')

        if calibrate_non_residential:

            valid_non_resid_elec = data_loader.read_lad_demands(
                data['paths']['val_subnational_elec_non_residential'])

            valid_non_resid_gas = data_loader.read_lad_demands(
                data['paths']['val_subnational_gas_non_residential'])

            # Calculate and apply regional calibration factor
            for region in data['regions']:

                # Total modeleld non_residential
                service_demand_elec = 0
                service_demand_gas = 0
                for enduse_data in disagg['service'][region].values():
                    for sector_data in enduse_data.values():
                        service_demand_elec += np.sum(sector_data[fueltype_elec])
                        service_demand_gas += np.sum(sector_data[fueltype_gas])

                industry_demand_elec = 0
                industry_demand_gas = 0
                for enduse_data in disagg['industry'][region].values():
                    for sector_data in enduse_data.values():
                        industry_demand_elec += np.sum(sector_data[fueltype_elec])
                        industry_demand_gas += np.sum(sector_data[fueltype_gas])

                modelled_elec = service_demand_elec + industry_demand_elec
                modelled_gas = service_demand_gas + industry_demand_gas

                # Calculate calibration factor
                try:
                    f_spatial_calibration_elec = valid_non_resid_elec[region] / modelled_elec
                except KeyError: # No real data available
                    f_spatial_calibration_elec = 1

                try:
                    f_spatial_calibration_gas = valid_non_resid_gas[region] / modelled_gas
                except KeyError: # No real data available
                    f_spatial_calibration_gas = 1

                # Apply calibration factor to spatial disaggregation
                for enduse in disagg['service'][region].keys():
                    for sector in disagg['service'][region][enduse]:
                        disagg['service'][region][enduse][sector][fueltype_elec] *= f_spatial_calibration_elec
                        disagg['service'][region][enduse][sector][fueltype_gas] *= f_spatial_calibration_gas
                for enduse in disagg['industry'][region].keys():
                    for sector in disagg['industry'][region][enduse]:
                        disagg['industry'][region][enduse][sector][fueltype_elec] *= f_spatial_calibration_elec
                        disagg['industry'][region][enduse][sector][fueltype_gas] *= f_spatial_calibration_gas

        if calibrate_residential:
            valid_resid_elec = data_loader.read_lad_demands(
                data['paths']['val_subnational_elec_residential'])
            valid_resid_gas = data_loader.read_lad_demands(
                data['paths']['val_subnational_gas_residential'])

            # Calculate and apply regional calibration factor
            for region in data['regions']:

                # Total modeleld residential
                modelled_elec = 0
                modelled_gas = 0
                for enduse_data in disagg['residential'][region].values():
                    modelled_elec += np.sum(enduse_data[fueltype_elec])
                    modelled_gas += np.sum(enduse_data[fueltype_gas])

                # Calculate calibration factor
                try:
                    real_elec = valid_resid_elec[region]
                    f_spatial_calibration_elec = real_elec / modelled_elec
                    real_gas = valid_resid_gas[region]
                    f_spatial_calibration_gas = real_gas / modelled_gas
                except KeyError:
                    # No real data available
                    f_spatial_calibration_elec = 1
                    f_spatial_calibration_gas = 1

                # Apply calibration factor to spatial disaggregation
                for enduse in disagg['residential'][region].keys():
                    disagg['residential'][region][enduse][fueltype_elec] *= f_spatial_calibration_elec
                    disagg['residential'][region][enduse][fueltype_gas] *= f_spatial_calibration_gas
    else:
        pass

    # Sum demand across all sectors for every region
    disagg['ss_fuel_disagg_sum_all_sectors'] = init_scripts.sum_across_sectors_all_regs(
        disagg['service'])

    disagg['is_aggr_fuel_sum_all_sectors'] = init_scripts.sum_across_sectors_all_regs(
        disagg['industry'])

    # Sum demand across all submodels and sectors for every region
    disagg['tot_disaggregated_regs'] = init_scripts.sum_across_all_submodels_regs(
        data['regions'],
        [disagg['residential'], disagg['service'], disagg['industry']])

    disagg['tot_disaggregated_regs_residenital'] = init_scripts.sum_across_all_submodels_regs(
        data['regions'],
        [disagg['residential']])

    disagg['tot_disaggregated_regs_non_residential'] = init_scripts.sum_across_all_submodels_regs(
        data['regions'],
        [disagg['service'], disagg['industry']])

    return disagg

def replace_msoa_reg_with_lad(region):
    """Replace mosa with corresponding LAD
    """
    lu_lad_msoa = lookup_tables.lad_msoa_mapping()

    for reg_lad, msoas_lad in lu_lad_msoa.items():
        if region in list(msoas_lad):
            #logging.info("Replaced msoa region %s with lad region %s", region, reg_lad)
            region = reg_lad
            break

    return region

def disaggregate_base_demand(
        pop_for_disagg,
        regions,
        fuels,
        scenario_data,
        assumptions,
        reg_coord,
        weather_stations,
        temp_data,
        sectors,
        enduses,
        service_building_count,
        crit_temp_min_max
    ):
    """This function disaggregates fuel demand based on
    region specific parameters for the base year. The residential,
    service and industry demand is disaggregated according to
    different factors.

    Arguments
    ----------
    pop_for_disagg : dict
        Population of years
    regions : list
        Regions
    fuels : dict
        Fuels
    scenario_data : dict
        Scenario data
    assumptions : obj
        Assumptions
    reg_coord : dict
        Region coordinates
    weather_stations : dict
        Weather stations
    temp_data : dict
        Temperature data
    sectors : dict
        Sectors per submodel
    enduses : dict
        Enduses per submodel
    service_building_count :

    Returns
    -------
    data : dict
    """
    # -------------------------------------
    # Factors to choose for disaggregation
    # -------------------------------------
    crit_limited_disagg_pop = False      # Only population
    crit_limited_disagg_pop_hdd = True   # Only population and HDD
    crit_full_disagg = False             # Use floor area as well
    census_disagg = True                 # Disagregatte is_demand with census statistics

    # Residential
    rs_fuel_disagg = rs_disaggregate(
        regions,
        fuels['residential'],
        scenario_data,
        pop_for_disagg,
        assumptions,
        reg_coord,
        weather_stations,
        temp_data,
        enduses['residential'],
        crit_limited_disagg_pop_hdd,
        crit_limited_disagg_pop,
        crit_full_disagg,
        crit_temp_min_max)

    # Service
    ss_fuel_disagg = ss_disaggregate(
        fuels['service'],
        service_building_count,
        assumptions,
        scenario_data,
        pop_for_disagg,
        regions,
        reg_coord,
        temp_data,
        weather_stations,
        enduses['service'],
        sectors['service'],
        crit_limited_disagg_pop_hdd,
        crit_limited_disagg_pop,
        crit_full_disagg,
        crit_temp_min_max)

    # Industry
    is_fuel_disagg = is_disaggregate(
        assumptions,
        temp_data,
        reg_coord,
        weather_stations,
        fuels['industry'],
        regions,
        enduses['industry'],
        sectors['industry'],
        scenario_data['employment_stats'],
        pop_for_disagg,
        census_disagg=census_disagg,
        crit_temp_min_max=crit_temp_min_max)

    return dict(rs_fuel_disagg), dict(ss_fuel_disagg), dict(is_fuel_disagg)

def ss_disaggregate(
        ss_national_fuel,
        service_building_count,
        assumptions,
        scenario_data,
        pop_for_disagg,
        regions,
        reg_coord,
        temp_data,
        weather_stations,
        enduses,
        sectors,
        crit_limited_disagg_pop_hdd,
        crit_limited_disagg_pop,
        crit_full_disagg,
        crit_temp_min_max=False
    ):
    """Disaggregate fuel for service submodel (per enduse and sector)

    Outputs
    -------
    ss_fuel_disagg : dict
        region, Enduse, Sectors
    """
    logging.debug("... disaggregate service demand")
    ss_fuel_disagg = {}

    # ---------------------------------------
    # Calculate heating degree days for regions
    # ---------------------------------------
    ss_hdd_individ_region = hdd_cdd.get_hdd_country(
        t_base_heating=assumptions.t_bases.ss_t_heating,
        regions=regions,
        temp_data=temp_data,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        crit_temp_min_max=crit_temp_min_max)

    ss_cdd_individ_region = hdd_cdd.get_cdd_country(
        t_base_cooling=assumptions.t_bases.ss_t_cooling,
        regions=regions,
        temp_data=temp_data,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        crit_temp_min_max=crit_temp_min_max)

    # ---------------------------------------------
    # Get all regions with floor area data
    # ---------------------------------------------
    regions_with_floorarea = list(regions)
    for reg in assumptions.ss_regions_without_floorarea:
        regions_with_floorarea.remove(reg)

    # ---------------------------------------
    # Overall disaggregation factors per enduse and sector
    # ---------------------------------------
    ss_fuel_disagg = {}
    ss_fuel_disagg_only_pop = ss_disaggr(
        all_regions=regions,
        regions=assumptions.ss_regions_without_floorarea,
        sectors=sectors,
        enduses=enduses,
        base_yr=assumptions.base_yr,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disagg,
        service_building_count=service_building_count,
        ss_hdd_individ_region=ss_hdd_individ_region,
        ss_cdd_individ_region=ss_cdd_individ_region,
        ss_national_fuel=ss_national_fuel,
        crit_limited_disagg_pop=False,      # Set to False
        crit_limited_disagg_pop_hdd=True,   # Set to True
        crit_full_disagg=False)
    ss_fuel_disagg.update(ss_fuel_disagg_only_pop)

    # Substract from national fuel already disaggregated fuel
    ss_national_fuel_remaining = copy.deepcopy(ss_national_fuel)
    for enduse in ss_national_fuel:
        for sector in ss_national_fuel[enduse]:
            for reg in ss_fuel_disagg:
                ss_national_fuel_remaining[enduse][sector] -= ss_fuel_disagg[reg][enduse][sector]

    # Disaggregate with floor area
    ss_fuel_disagg_full_data = ss_disaggr(
        all_regions=regions_with_floorarea,
        regions=regions_with_floorarea,
        sectors=sectors,
        enduses=enduses,
        base_yr=assumptions.base_yr,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disagg,
        service_building_count=service_building_count,
        ss_hdd_individ_region=ss_hdd_individ_region,
        ss_cdd_individ_region=ss_cdd_individ_region,
        ss_national_fuel=ss_national_fuel_remaining,
        crit_limited_disagg_pop=crit_limited_disagg_pop,
        crit_limited_disagg_pop_hdd=crit_limited_disagg_pop_hdd,
        crit_full_disagg=crit_full_disagg)
    ss_fuel_disagg.update(ss_fuel_disagg_full_data)

    # -----------------
    # Check if total fuel is the same before and after aggregation
    #------------------
    testing_functions.control_disaggregation(
        ss_fuel_disagg, ss_national_fuel, enduses, sectors)

    return dict(ss_fuel_disagg)

def ss_disaggr(
        all_regions,
        regions,
        sectors,
        enduses,
        base_yr,
        scenario_data,
        pop_for_disagg,
        service_building_count,
        ss_hdd_individ_region,
        ss_cdd_individ_region,
        ss_national_fuel,
        crit_limited_disagg_pop,
        crit_limited_disagg_pop_hdd,
        crit_full_disagg
    ):
    """Disaggregating
    """
    ss_fuel_disagg = {}
    tot_pop = 0
    tot_cdd = 0
    tot_hdd = 0
    tot_floor_area = {}
    tot_floor_area_pop = {}
    tot_pop_hdd = {}
    tot_pop_cdd = {}
    tot_floor_area_hdd = {}
    tot_floor_area_cdd = {}
    tot_building_cnt = {}

    lu_tables = lookup_tables.basic_lookups()
    lu_building_cnt = lu_tables['building_cnt_lu']

    for sector in sectors:
        tot_floor_area[sector] = 0
        tot_floor_area_pop[sector] = 0
        tot_pop_hdd[sector] = 0
        tot_pop_cdd[sector] = 0
        tot_floor_area_hdd[sector] = 0
        tot_floor_area_cdd[sector] = 0
        tot_building_cnt[sector] = 0

    for region in all_regions:
        reg_hdd = ss_hdd_individ_region[region]
        reg_cdd = ss_cdd_individ_region[region]
        reg_pop = pop_for_disagg[base_yr][region]
        tot_pop += reg_pop
        tot_cdd += reg_cdd
        tot_hdd += reg_hdd

        for sector in sectors:

            # Sector specific ino
            reg_floor_area = scenario_data['floor_area']['ss_floorarea'][base_yr][region][sector]

            try:
                nr_sector_cnt_building = lu_building_cnt[sector]
                reg_sector_building_cnt = service_building_count[nr_sector_cnt_building][region]
                tot_building_cnt[sector] += reg_sector_building_cnt
            except KeyError:
                #logging.debug("No building data for region {}".format(region))
                pass

            # National disaggregation factors
            tot_floor_area[sector] += reg_floor_area
            tot_floor_area_pop[sector] += reg_floor_area * reg_pop
            tot_floor_area_hdd[sector] += reg_floor_area * reg_hdd
            tot_pop_hdd[sector] += reg_pop * reg_hdd
            tot_pop_cdd[sector] += reg_pop * reg_cdd
            tot_floor_area_cdd[sector] += reg_floor_area * reg_cdd

    # ---------------------------------------
    # Disaggregate according to enduse
    # ---------------------------------------
    for region in regions:
        ss_fuel_disagg[region] = defaultdict(dict)

        # Regional factors
        reg_hdd = ss_hdd_individ_region[region]
        reg_cdd = ss_cdd_individ_region[region]
        reg_pop = pop_for_disagg[base_yr][region]
        p_pop = reg_pop / tot_pop

        for enduse in enduses:
            for sector in sectors:
                if crit_limited_disagg_pop and not crit_limited_disagg_pop_hdd and not crit_full_disagg:

                    # ----
                    #logging.debug(" ... Disaggregation ss: populaton")
                    # ----
                    reg_diasg_factor = p_pop

                elif crit_limited_disagg_pop_hdd and not crit_full_disagg:

                    # ----
                    #logging.debug(" ... Disaggregation ss: populaton, HDD")
                    # ----
                    if enduse == 'ss_cooling_humidification':
                        reg_diasg_factor = (reg_pop * reg_cdd) / tot_pop_cdd[sector]

                    elif enduse == 'ss_space_heating':
                        reg_diasg_factor = (reg_pop * reg_hdd) / tot_pop_hdd[sector]
                    else:
                        reg_diasg_factor = p_pop

                elif crit_full_disagg:

                    reg_floor_area = scenario_data['floor_area']['ss_floorarea'][base_yr][region][sector]

                    # ----
                    # logging.debug(" ... Disaggregation ss: populaton, HDD, floor_area")
                    # ----
                    if enduse == 'ss_cooling_humidification':
                        reg_diasg_factor = (reg_floor_area * reg_cdd) / tot_floor_area_cdd[sector]
                    elif enduse == 'ss_space_heating':
                        reg_diasg_factor = (reg_floor_area * reg_hdd) / tot_floor_area_hdd[sector]
                    elif enduse == 'ss_lighting':
                        reg_diasg_factor = reg_floor_area / tot_floor_area[sector]
                    else:

                        # Disaggregate with population
                        #reg_diasg_factor = p_pop

                        # Disaggregate with nr of buildings
                        nr_sector_cnt_building = lu_building_cnt[sector]
                        reg_sector_building_cnt = service_building_count[nr_sector_cnt_building][region]
                        p_reg_sector_building_cnt = reg_sector_building_cnt / tot_building_cnt[sector]

                        reg_diasg_factor = p_reg_sector_building_cnt

                # If reg_diasg_factor, replace by population disaggregation
                if np.isnan(reg_diasg_factor):
                    reg_diasg_factor = p_pop
                else:
                    pass

                ss_fuel_disagg[region][enduse][sector] = ss_national_fuel[enduse][sector] * reg_diasg_factor

    return dict(ss_fuel_disagg)

def is_disaggregate(
        assumptions,
        temp_data,
        reg_coord,
        weather_stations,
        is_national_fuel,
        regions,
        enduses,
        sectors,
        employment_statistics,
        pop_for_disagg,
        census_disagg,
        crit_temp_min_max=False
    ):
    """Disaggregate industry related fuel for sector and enduses with
    employment statistics

    base_yr : int
        Base year
    is_national_fuel : dict
        Fuel
    regions : list
        Regions
    enduses : list
        Enduses
    sectors : list
        Sectors
    employment_statistics : dict
        Employment statistics
    crit_limited_disagg_pop : bool
        Criteria which diassgregation method
    crit_full_disagg : bool
        Criteria which diassgregation method

    Returns
    ---------
    is_fuel_disagg : dict
        reg, enduse, sector
    """
    is_hdd_individ_region = hdd_cdd.get_hdd_country(
        t_base_heating=assumptions.t_bases.is_t_heating,
        regions=regions,
        temp_data=temp_data,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        crit_temp_min_max=crit_temp_min_max)

    is_fuel_disagg = {}

    # # Calculate total population of all regions
    tot_pop = 0
    tot_pop_hdd = 0
    for region in regions:
        tot_pop_hdd += pop_for_disagg[assumptions.base_yr][region] * is_hdd_individ_region[region]
        tot_pop += pop_for_disagg[assumptions.base_yr][region]

    if not census_disagg:

        # ---
        # Disaggregate only with population
        # ---
        for region in regions:
            is_fuel_disagg[region] = {}

            reg_pop = pop_for_disagg[assumptions.base_yr][region]

            for enduse in enduses:
                is_fuel_disagg[region][enduse] = {}

                for sector in sectors:
                    if enduse == 'is_space_heating':
                        hdd_reg = is_hdd_individ_region[region]
                        reg_disagg_f = (hdd_reg * reg_pop) / tot_pop_hdd
                        is_fuel_disagg[region][enduse][sector] = is_national_fuel[enduse][sector] * reg_disagg_f
                    else:
                        reg_disagg_f = reg_pop / tot_pop
                        is_fuel_disagg[region][enduse][sector] = is_national_fuel[enduse][sector] * reg_disagg_f

        return is_fuel_disagg

    else:
        #logging.debug(" ... Disaggregation is: Employment statistics")
        # -----
        # Disaggregate with employment statistics
        # The BEIS sectors are matched with census data sectors {ECUK industry sectors: 'Emplyoment sectors'}
        sectormatch_ecuk_with_census = {

            # Significant improvement
            'mining': 'B', #'B',
            'food_production': 'C10-12',
            'pharmaceuticals': 'M',
            'computer': 'C26-30',
            'leather': 'C13-15',
            'wearing_appeal': 'C13-15',

            # Improvement
            'basic_metals': 'C',
            'non_metallic_mineral_products': 'C',
            'electrical_equipment': 'C26-30',
            'printing': 'C',
            'rubber_plastics': 'C19-22',
            'chemicals': 'C19-22',
            'wood': 'C16,17',
            'paper': 'C16,17',

            # Worse and better
            'fabricated_metal_products': 'C',   # Gas better, elec worse test C23-25  previous 'C'
            'textiles': 'C13-15',               # Gas better, elec worse
            'motor_vehicles': 'G',              # Gas better, elec worse

            # Unclear
            'machinery': None,                  # 'C'
            'tobacco': None,                    # 'C10-12'
            'other_transport_equipment': None,  # 'H'
            'other_manufacturing': None,        # 'C18,31,32'
            'water_collection_treatment': None, # 'E'
            'waste_collection': None,           # 'E'
            'furniture': None,                  # C18,31,32
            'beverages': None
        }
        # ----------------------------------------
        # Summarise national employment per sector
        # ----------------------------------------
        tot_national_sector_employment = {}
        tot_national_sector_employment_hdd = {}
        for region in regions:
            is_hdd = is_hdd_individ_region[region]

            # -------------------------------------------------------------------
            # In case MSOA are provided as input, the corresponing LAD is searched
            # -------------------------------------------------------------------
            region_remapped = replace_msoa_reg_with_lad(region)

            for employment_sector, employment in employment_statistics[region_remapped].items():
                try:
                    tot_national_sector_employment[employment_sector] += employment
                    tot_national_sector_employment_hdd[employment_sector] += employment * is_hdd
                except KeyError:
                    tot_national_sector_employment[employment_sector] = employment
                    tot_national_sector_employment_hdd[employment_sector] = employment * is_hdd

        # --------------------------------------------------
        # Disaggregate per region with employment statistics
        # --------------------------------------------------
        for region in regions:
            is_fuel_disagg[region] = {}
            reg_pop = pop_for_disagg[assumptions.base_yr][region]

            for enduse in enduses:
                is_fuel_disagg[region][enduse] = {}

                for sector in sectors:

                    # ---------------------------------
                    # Try to match  with sector, otherwise disaggregate with population
                    # ----------------------------------
                    matched_sector = sectormatch_ecuk_with_census[sector]

                    # Disaggregate with population
                    if not matched_sector:

                        if enduse == 'is_space_heating':
                            hdd_reg = is_hdd_individ_region[region]
                            reg_disagg_f = (hdd_reg * reg_pop) / tot_pop_hdd
                            is_fuel_disagg[region][enduse][sector] = is_national_fuel[enduse][sector] * reg_disagg_f
                        else:
                            reg_disagg_f = reg_pop / tot_pop
                            is_fuel_disagg[region][enduse][sector] = is_national_fuel[enduse][sector] * reg_disagg_f
                    else:
                        # -------------------------------------------------------------------
                        # In case MSOA are provided as input, the corresponing LAD is searched
                        # -------------------------------------------------------------------
                        region_remapped = replace_msoa_reg_with_lad(region)

                        reg_sector_employment = employment_statistics[region_remapped][matched_sector]

                        if enduse == 'is_space_heating':
                            national_sector_employment_hdd = tot_national_sector_employment_hdd[matched_sector]
                            hdd_reg = is_hdd_individ_region[region]
                            reg_disagg_f = (hdd_reg * reg_sector_employment) / national_sector_employment_hdd
                        else:
                            national_sector_employment = tot_national_sector_employment[matched_sector]
                            try:
                                reg_disagg_f = reg_sector_employment / national_sector_employment
                            except ZeroDivisionError:
                                reg_disagg_f = 0 #No employment for this sector for this region

                        # If nan, use population
                        if np.isnan(reg_disagg_f):
                            reg_disagg_f = reg_pop / tot_pop
                        else:
                            pass

                        # Disaggregated national fuel
                        is_fuel_disagg[region][enduse][sector] = is_national_fuel[enduse][sector] * reg_disagg_f

    testing_functions.control_disaggregation(
        is_fuel_disagg,
        is_national_fuel,
        enduses,
        sectors)

    logging.debug("... finished disaggregateing industry demand")
    return is_fuel_disagg

def rs_disaggregate(
        regions,
        rs_national_fuel,
        scenario_data,
        pop_for_disagg,
        assumptions,
        reg_coord,
        weather_stations,
        temp_data,
        enduses,
        crit_limited_disagg_pop_hdd,
        crit_limited_disagg_pop,
        crit_full_disagg,
        crit_temp_min_max=False,
        dummy_sector=None
    ):
    """Disaggregate residential fuel demand

    Arguments
    ----------
    regions : dict
        Regions
    rs_national_fuel : dict
        Fuel per enduse for residential submodel

    Returns
    -------
    rs_fuel_disagg : dict
        Disaggregated fuel per enduse for every region (fuel[region][enduse])

    Note
    -----
    Used disaggregation factors for residential according
    to enduse (see Documentation)
    """
    logging.debug("... disagreggate residential demand")

    rs_fuel_disagg = {}
    # ---------------------------------------
    # Calculate heating degree days for regions
    # ---------------------------------------
    rs_hdd_individ_region = hdd_cdd.get_hdd_country(
        t_base_heating=assumptions.t_bases.rs_t_heating,
        regions=regions,
        temp_data=temp_data,
        reg_coord=reg_coord,
        weather_stations=weather_stations,
        crit_temp_min_max=crit_temp_min_max)

    # ---------------------------------------
    # Get all regions with  floor area data
    # ---------------------------------------
    regions_with_floorarea = list(regions)
    for reg in assumptions.rs_regions_without_floorarea:
        regions_with_floorarea.remove(reg)

    # ====================================
    # Disaggregate for region without floor
    # area by population and hdd (set crit_full_disagg as False)
    # ====================================
    rs_fuel_regions = rs_disaggr(
        all_regions=regions,
        regions=assumptions.rs_regions_without_floorarea,
        base_yr=assumptions.base_yr,
        rs_hdd_individ_region=rs_hdd_individ_region,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disagg,
        rs_national_fuel=rs_national_fuel,
        crit_limited_disagg_pop=False, #True, #False,
        crit_limited_disagg_pop_hdd=True, #True, #Set to true
        crit_full_disagg=False,
        dummy_sector=dummy_sector)
    rs_fuel_disagg.update(rs_fuel_regions)

    # Substract fuel for regions where only population was used for disaggregation from total
    rs_national_fuel_remaining = copy.deepcopy(rs_national_fuel)
    for enduse in rs_national_fuel:
        for reg in rs_fuel_disagg:
            rs_national_fuel_remaining[enduse][dummy_sector] -= rs_fuel_disagg[reg][enduse][dummy_sector]

    # ====================================
    # Disaggregate for region with floor area
    # ====================================
    rs_fuel_disagg_full_data = rs_disaggr(
        all_regions=regions_with_floorarea,
        regions=regions_with_floorarea,
        base_yr=assumptions.base_yr,
        rs_hdd_individ_region=rs_hdd_individ_region,
        scenario_data=scenario_data,
        pop_for_disagg=pop_for_disagg,
        rs_national_fuel=rs_national_fuel_remaining,
        crit_limited_disagg_pop=crit_limited_disagg_pop,
        crit_limited_disagg_pop_hdd=crit_limited_disagg_pop_hdd,
        crit_full_disagg=crit_full_disagg,
        dummy_sector=dummy_sector)
    rs_fuel_disagg.update(rs_fuel_disagg_full_data)

    # Check if total fuel is the same before and after aggregation
    testing_functions.control_disaggregation(
        rs_fuel_disagg, rs_national_fuel, enduses, [dummy_sector])

    return rs_fuel_disagg

def rs_disaggr(
        all_regions,
        regions,
        base_yr,
        rs_hdd_individ_region,
        scenario_data,
        pop_for_disagg,
        rs_national_fuel,
        crit_limited_disagg_pop,
        crit_limited_disagg_pop_hdd,
        crit_full_disagg,
        dummy_sector
    ):
    """Disaggregate residential enduses
    """
    fuel_disagg = {}

    total_pop = 0
    total_hdd_floorarea = 0
    total_floor_area = 0
    total_pop_hdd = 0
    total_hdd = 0

    for region in all_regions:
        reg_hdd = rs_hdd_individ_region[region]
        reg_floor_area = scenario_data['floor_area']['rs_floorarea'][base_yr][region]
        reg_pop = pop_for_disagg[base_yr][region]

        # National dissagregation factors
        total_hdd_floorarea += reg_hdd * reg_floor_area
        total_floor_area += reg_floor_area
        total_pop += reg_pop
        total_pop_hdd += reg_pop * reg_hdd
        total_hdd += reg_hdd

    # ---------------------------------------
    # Disaggregate according to enduse
    # ---------------------------------------
    for region in regions:
        fuel_disagg[region] = {}
        reg_hdd = rs_hdd_individ_region[region]
        reg_floor_area = scenario_data['floor_area']['rs_floorarea'][base_yr][region]
        reg_hdd_floor_area = reg_hdd * reg_floor_area
        reg_pop = pop_for_disagg[base_yr][region]
        reg_pop_hdd = reg_pop * reg_hdd

        p_pop = reg_pop / total_pop
        p_floor_area = reg_floor_area / total_floor_area

        # Disaggregate fuel depending on end_use
        for enduse in rs_national_fuel:
            fuel_disagg[region][enduse] = {}
            if crit_limited_disagg_pop and not crit_limited_disagg_pop_hdd and not crit_full_disagg:

                # ----------------------------------
                #logging.debug(" ... Disaggregation rss: populaton")
                # ----------------------------------
                reg_diasg_factor = p_pop

            elif crit_limited_disagg_pop_hdd and not crit_full_disagg:

                # -------------------
                #logging.debug(" ... Disaggregation rss: populaton, hdd")
                # -------------------
                if enduse == 'rs_space_heating':
                    reg_diasg_factor = reg_pop_hdd / total_pop_hdd
                else:
                    reg_diasg_factor = p_pop

            elif crit_full_disagg:

                # -------------------
                #logging.debug(" ... Disaggregation rss: populaton, hdd, floor_area")
                # -------------------
                if enduse == 'rs_space_heating':
                    reg_diasg_factor = reg_hdd_floor_area / total_hdd_floorarea
                elif enduse == 'rs_lighting':
                    reg_diasg_factor = p_floor_area
                else:
                    reg_diasg_factor = p_pop

            if np.isnan(reg_diasg_factor):
                reg_diasg_factor = p_pop
            else:
                pass

            # Disaggregate
            fuel_disagg[region][enduse][dummy_sector] = rs_national_fuel[enduse][dummy_sector] * reg_diasg_factor

    return dict(fuel_disagg)

def write_disagg_fuel(path_to_txt, data):
    """Write out disaggregated fuel

    Arguments
    ----------
    path_to_txt : str
        Path to txt file
    data : dict
        Data to write out
    """
    file = open(path_to_txt, "w")
    file.write("{}, {}, {}, {}".format(
        'region', 'enduse', 'fueltypes', 'fuel') + '\n')

    for region, enduses in data.items():
        for enduse, fuels in enduses.items():
            for fueltype, fuel in enumerate(fuels):
                file.write("{}, {}, {}, {}".format(
                    str.strip(region),
                    str.strip(enduse),
                    str(int(fueltype)),
                    str(float(fuel)) + '\n'))
    file.close()

    return

def write_disagg_fuel_ts(path_to_txt, data):
    """Write out disaggregated fuel

    Arguments
    ----------
    path_to_txt : str
        Path to txt file
    data : dict
        Data to write out
    """
    file = open(path_to_txt, "w")
    file.write("{}, {}, {}".format(
        'region', 'fueltypes', 'fuel') + '\n')

    for region, fuels in data.items():
        for fueltype, fuel in enumerate(fuels):
            file.write("{}, {}, {}".format(
                str.strip(region), str(int(fueltype)), str(float(fuel)) + '\n')
                      )
    file.close()

    return

def write_disagg_fuel_sector(path_to_txt, data):
    """Write out disaggregated fuel

    Arguments
    ----------
    path_to_txt : str
        Path to txt file
    data : dict
        Data to write out
    """
    file = open(path_to_txt, "w")
    file.write("{}, {}, {}, {}, {}".format(
        'region', 'enduse', 'sector', 'fueltypes', 'fuel') + '\n')

    for region, sectors in data.items():
        for sector, enduses in sectors.items():
            for enduse, fuels in enduses.items():
                for fueltype, fuel in enumerate(fuels):
                    file.write("{}, {}, {}, {}, {}".format(
                        str.strip(region),
                        str.strip(enduse),
                        str.strip(sector),
                        str(int(fueltype)),
                        str(float(fuel)) + '\n'))
    file.close()

    return



"""
"""
import numpy as np
import matplotlib.pyplot as plt

from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import plotting_styles
from energy_demand.basic import conversions

def run(results, lookups, fig_name, plotshow=False):
    """Plot lines with total energy demand for all enduses
    per fueltype over the simluation period. Annual GWh
    are converted into GW.

    Arguments
    ---------
    results : dict
        Results for every year and fueltype (yh)
    lookups : dict
        Lookup fueltypes
    fig_name : str
        Figure name

    Note
    ----
    Values are divided by 1'000
    """
    print("... plot fuel per fueltype for whole country over annual timesteps")

    # Set figure size
    plt.figure(figsize=basic_plot_functions.cm2inch(14, 8))

    # Initialise (number of enduses, number of hours to plot)
    y_values_fueltype = {}

    for fueltype_str, fueltype_int in lookups['fueltypes'].items():

        # Read out fueltype specific max h load
        data_years = {}
        for year, data_year in results.items():
            tot_gwh_fueltype_y = np.sum(data_year[fueltype_int])

            #Conversion: Convert gwh per years to gw
            yearly_sum_gw = tot_gwh_fueltype_y

            yearly_sum_twh = conversions.gwh_to_twh(yearly_sum_gw)

            data_years[year] = yearly_sum_twh #yearly_sum_gw

        y_values_fueltype[fueltype_str] = data_years

    # -----------------
    # Axis
    # -----------------
    base_yr, year_interval = 2015, 5
    end_yr = list(results.keys())

    major_ticks = np.arange(
        base_yr,
        end_yr[-1] + year_interval,
        year_interval)

    plt.xticks(major_ticks, major_ticks)

    # ----------
    # Plot lines
    # ----------
    color_list_selection = plotting_styles.color_list_selection()

    for fueltype_str, fuel_fueltype_yrs in y_values_fueltype.items():

        if len(np.array(list(fuel_fueltype_yrs.values()))) > 2:
            smooth_x_line_data, smooth_y_line_data = basic_plot_functions.smooth_line(
                np.array(list(fuel_fueltype_yrs.keys())),
                np.array(list(fuel_fueltype_yrs.values())))
        else:
            smooth_x_line_data = list(fuel_fueltype_yrs.keys())
            smooth_y_line_data = list(fuel_fueltype_yrs.values())

        plt.plot(
            smooth_x_line_data,     # years
            smooth_y_line_data,   # yearly data per fueltype
            color=str(color_list_selection.pop()),
            label=fueltype_str)

    # ----
    # Axis
    # ----
    plt.ylim(ymin=0) #no upper limit to xmax

    # ------------
    # Plot legend
    # ------------
    plt.legend(
        ncol=2,
        loc=2,
        prop={
            'family': 'arial',
            'size': 10},
        frameon=False)

    # ---------
    # Labels
    # ---------
    plt.ylabel("TWh")
    plt.xlabel("year")
    plt.title("tot annual ED per fueltype")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.savefig(fig_name)

    if plotshow:
        plt.show()
        plt.close()
    else:
        plt.close()
    


"""Read in model results and plot results
"""
import os
import logging
import numpy as np
import time

from energy_demand.read_write import data_loader, read_data
from energy_demand.technologies import tech_related

def read_in_weather_results(
        path_result,
        seasons,
        model_yeardays_daytype,
        pop_data,
        fueltype_str
    ):
    """Read and post calculate results from txt files
    and store into container

    Arguments
    ---------
    path_result : str
        Paths
    seasons : dict
        seasons
    model_yeardays_daytype : dict
        Daytype of modelled yeardays
    """
    logging.info("... Reading in results")

    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    results_container = {}

    # Read in total regional demands per fueltype
    results_container['ed_reg_tot_y'] = read_data.read_results_yh(
        path_result, 'only_total')

    #print(results_container['ed_reg_tot_y'][2015].shape)
    results_container['ed_reg_peakday'] = read_data.read_results_yh(
        os.path.join('simulation_results', path_result), 'only_peak')

    #print(results_container['ed_reg_peakday'][2015].shape)
    results_container['ed_reg_peakday_peak_hour'] = {}
    results_container['ed_reg_peakday_peak_hour_per_pop'] = {}
    
    results_container['national_peak'] = {}
    results_container['regional_share_national_peak'] = {}
    results_container['regional_share_national_peak_pp'] = {}
    results_container['pp_peak_abs'] = {}
    results_container['regional_peak'] = {}
    results_container['national_all_fueltypes'] = {}

    for year in results_container['ed_reg_peakday']:

        reg_pop_yr = pop_data[year]

        # Get peak demand of each region
        results_container['ed_reg_peakday_peak_hour'][year] = results_container['ed_reg_peakday'][year].max(axis=2)

        # Divide peak by number of population
        results_container['ed_reg_peakday_peak_hour_per_pop'][year] = results_container['ed_reg_peakday_peak_hour'][year] / reg_pop_yr

        # Get national peak
        national_demand_per_hour = results_container['ed_reg_peakday'][year].sum(axis=1) #Aggregate hourly across all regions

        # Get maximum hour for electricity demand
        max_hour = national_demand_per_hour[fueltype_int].argmax()

        results_container['national_peak'][year] = national_demand_per_hour[:, max_hour]

        # Calculate regional share of peak hour to national peak
        national_peak = results_container['national_peak'][year][fueltype_int]
        regional_peak = results_container['ed_reg_peakday'][year][fueltype_int][:, max_hour]
        
        results_container['regional_peak'][year] = regional_peak
        results_container['regional_share_national_peak'][year] = (100 / national_peak) * regional_peak #1 = 1 %
        results_container['regional_share_national_peak_pp'][year] = ((100 / national_peak) * regional_peak) /  reg_pop_yr #1 = 1 %

        # Sum all regions for each fueltypes
        #print(results_container['ed_reg_tot_y'][year].shape)
        results_container['national_all_fueltypes'][year] = np.sum(results_container['ed_reg_tot_y'][year], axis=1)
        #print(results_container['national_all_fueltypes'][year].shape)
    
        # Calculate contribution per person towards national peak (reg_peak / people) [abs]
        #print(results_container['ed_reg_peakday'][year].shape)
        #print(reg_pop_yr.shape)
        # results_container['pp_peak_abs'][year] = (
        #    results_container['ed_reg_peakday'][year][:,:, max_hour] / reg_pop_yr)

        #(cpp = (regional peak / national peak) / people [%]

    logging.info("... Reading in results finished")
    return results_container


"""
"""
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import geopandas as gpd

from energy_demand.read_write import data_loader, read_data
from energy_demand.basic import date_prop
from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables
from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import result_mapping
from energy_demand.plotting import fig_p2_weather_val

def run_fig_spatial_distribution_of_peak(
        scenarios,
        path_to_folder_with_scenarios,
        path_shapefile,
        sim_yrs,
        field_to_plot,
        unit,
        fig_path,
        fueltype_str='electricity'
    ):
    """
    """
    weather_yrs = []
    calculated_yrs_paths = []
    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    for scenario in scenarios:
        path_scenario = os.path.join(path_to_folder_with_scenarios, scenario)
        all_result_folders = os.listdir(path_scenario)

        for result_folder in all_result_folders:
            try:
                split_path_name = result_folder.split("__")
                weather_yr = int(split_path_name[0])
                weather_yrs.append(weather_yr)
                tupyle_yr_path = (weather_yr, os.path.join(path_scenario))
                calculated_yrs_paths.append(tupyle_yr_path)
            except ValueError:
                pass

    for simulation_yr in sim_yrs:
        container = {}
        container['abs_demand_in_peak_h_pp'] = {}
        container['abs_demand_in_peak_h'] = {}
        container['p_demand_in_peak_h'] = {}

        for weather_yr, path_data_ed in calculated_yrs_paths:
            print("... prepare data {} {}".format(weather_yr, path_data_ed))

            path_to_weather_yr = os.path.join(path_data_ed, "{}__{}".format(weather_yr, 'all_stations'))

            data = {}
            data['lookups'] = lookup_tables.basic_lookups()
            data['enduses'], data['assumptions'], regions = data_loader.load_ini_param(os.path.join(path_data_ed))
            data['assumptions']['seasons'] = date_prop.get_season(year_to_model=2015)
            data['assumptions']['model_yeardays_daytype'], data['assumptions']['yeardays_month'], data['assumptions']['yeardays_month_days'] = date_prop.get_yeardays_daytype(year_to_model=2015)

            # Population 
            population_data = read_data.read_scenaric_population_data(os.path.join(path_data_ed, 'model_run_pop'))

            results_container = read_data.read_in_results(
                os.path.join(path_to_weather_yr, 'model_run_results_txt'),
                data['assumptions']['seasons'],
                data['assumptions']['model_yeardays_daytype'])

            # ---------------------------------------------------
            # Calculate hour with national peak demand
            # This may be different depending on the weather yr
            # ---------------------------------------------------
            ele_regions_8760 = results_container['ed_fueltype_regs_yh'][simulation_yr][fueltype_int]
            sum_all_regs_fueltype_8760 = np.sum(ele_regions_8760, axis=0) # Sum for every hour

            max_day = int(basic_functions.round_down((np.argmax(sum_all_regs_fueltype_8760) / 24), 1))
            max_h = np.argmax(sum_all_regs_fueltype_8760)
            max_demand = np.max(sum_all_regs_fueltype_8760)

            # Calculate the national peak demand in GW
            national_peak_GW = np.max(sum_all_regs_fueltype_8760)

            # ------------------------------------------------------
            # Calculate the contribution of the regional peak demand
            # ------------------------------------------------------
            demand_in_peak_h = ele_regions_8760[:, max_h]

            if unit == 'GW':
                container['abs_demand_in_peak_h'][weather_yr] = demand_in_peak_h
            elif unit == 'kW':
                container['abs_demand_in_peak_h'][weather_yr] = demand_in_peak_h * 1000000 # Convert to KWh
            else:
                # Use GW as default
                container['abs_demand_in_peak_h'][weather_yr] = demand_in_peak_h

            container['abs_demand_in_peak_h_pp'][weather_yr] = demand_in_peak_h / population_data[simulation_yr]
    
            # Relative fraction of regional demand in relation to peak
            container['p_demand_in_peak_h'][weather_yr] = (demand_in_peak_h / national_peak_GW ) * 100 # given as percent

            print("=================================")
            print("{}  {}  {}  {}".format(
                simulation_yr,
                weather_yr,
                np.sum(ele_regions_8760),
                national_peak_GW))

        # --------------
        # Create dataframe with all weather yrs calculatiosn for every region
            
        #               region1, region2, region3
        # weather yr1
        # weather yr2
        # --------------
        # Convert regional data to dataframe
        abs_demand_in_peak_h_pp = np.array(list(container['abs_demand_in_peak_h_pp'].values()))
        abs_demand_peak_h = np.array(list(container['abs_demand_in_peak_h'].values()))
        p_demand_peak_h = np.array(list(container['p_demand_in_peak_h'].values()))

        # Absolute demand
        df_abs_peak_demand = pd.DataFrame(
            abs_demand_peak_h,
            columns=regions,
            index=list(container['abs_demand_in_peak_h'].keys()))

        # Relative demand
        df_p_peak_demand = pd.DataFrame(
            p_demand_peak_h,
            columns=regions,
            index=list(container['p_demand_in_peak_h'].keys()))
        
        df_abs_demand_in_peak_h_pp = pd.DataFrame(
            abs_demand_in_peak_h_pp,
            columns=regions,
            index=list(container['abs_demand_in_peak_h_pp'].keys()))

        # Absolute peak value - mean
        max_peak_h_across_weather_yrs = df_abs_peak_demand.max()
        average_across_weather_yrs = df_abs_peak_demand.mean()
        diff_peak_h_minus_mean = max_peak_h_across_weather_yrs - average_across_weather_yrs

        for index, row in df_p_peak_demand.iterrows():
            print("Weather yr: {} Total p: {}".format(index, np.sum(row)))
            assert round(np.sum(row), 4) == 100.0

        # ----------------------------
        # Calculate standard deviation
        # ----------------------------
        std_deviation_df_abs_demand_in_peak_h_pp = df_abs_demand_in_peak_h_pp.std()
        std_deviation_abs_demand_peak_h = df_abs_peak_demand.std()
        std_deviation_p_demand_peak_h = df_p_peak_demand.std()

        print("=========")
        print("National stats")
        print("=========")
        print("Sum of std of absolut peak demand:  " + str(np.sum(std_deviation_abs_demand_peak_h)))

        # --------------------
        # Create map
        # --------------------
        regional_statistics_columns = [
            'name',
            'std_deviation_p_demand_peak_h',
            'std_deviation_abs_demand_peak_h',
            'std_deviation_df_abs_demand_in_peak_h_pp',
            'diff_peak_h_minus_mean']

        df_stats = pd.DataFrame(columns=regional_statistics_columns)

        for region_name in regions:

            # 'name', 'absolute_GW', 'p_GW_peak'
            line_entry = [[
                region_name,
                std_deviation_p_demand_peak_h[region_name],
                std_deviation_abs_demand_peak_h[region_name],
                std_deviation_df_abs_demand_in_peak_h_pp[region_name],
                diff_peak_h_minus_mean[region_name],
                ]]

            line_df = pd.DataFrame(line_entry, columns=regional_statistics_columns)
            df_stats = df_stats.append(line_df)

        # Load uk shapefile
        uk_shapefile = gpd.read_file(path_shapefile)

        # Merge stats to geopanda
        shp_gdp_merged = uk_shapefile.merge(
            df_stats,
            on='name')

        # Assign projection
        crs = {'init': 'epsg:27700'} #27700: OSGB_1936_British_National_Grid
        uk_gdf = gpd.GeoDataFrame(shp_gdp_merged, crs=crs)

        ax = uk_gdf.plot(
            figsize=basic_plot_functions.cm2inch(25, 20))

        nr_of_intervals = 6

        bin_values = result_mapping.get_reasonable_bin_values_II(
            data_to_plot=list(uk_gdf[field_to_plot]),
            nr_of_intervals=nr_of_intervals)
        # Maual bins
        #bin_values = [0, 0.005, 0.01, 0.015, 0.02, 0.025, 0.03]
        print(float(uk_gdf[field_to_plot].max()))
        print("BINS " + str(bin_values))

        uk_gdf, cmap_rgb_colors, color_zero, min_value, max_value = fig_p2_weather_val.user_defined_bin_classification(
            uk_gdf,
            field_to_plot,
            bin_values=bin_values)

        # plot with face color attribute
        uk_gdf.plot(
            ax=ax,
            facecolor=uk_gdf['bin_color'],
            edgecolor='black',
            linewidth=0.5)

        legend_handles = result_mapping.add_simple_legend(
            bin_values,
            cmap_rgb_colors,
            color_zero)

        plt.legend(
            handles=legend_handles,
            title="{}  [{}]".format(field_to_plot, unit),
            prop={'size': 8},
            #loc='upper center', bbox_to_anchor=(0.5, -0.05),
            loc='center left', bbox_to_anchor=(1, 0.5),
            frameon=False)

        # PLot bins on plot
        '''plt.text(
            -20,
            -20,
            bin_values[:-1], #leave away maximum value
            fontsize=8)'''

        plt.tight_layout()

        fig_out_path = os.path.join(fig_path, str(field_to_plot) + "__" + str(simulation_yr) + ".pdf")
        plt.savefig(fig_out_path)



"""All assumptions are either loaded in this file or definied here
"""
from energy_demand.read_write import read_data
from energy_demand.technologies import tech_related
from energy_demand.basic import testing_functions, date_prop
from energy_demand.assumptions import fuel_shares
from energy_demand.initalisations import helpers
from energy_demand.profiles import hdd_cdd
from energy_demand.read_write import narrative_related
from energy_demand.basic import lookup_tables

class Assumptions(object):
    """Assumptions of energy demand model

    Arguments
    ---------
    base_yr : int, default=None
        Base year
    curr_yr : int, default=None
        Current year
    sim_yrs : list, default=None
        Simulated years
    paths : dict, default=None
        Paths
    enduses : list, default=None
        All modelled end uses
    sectors : list, default=None
        All modelled sectors
    fueltypes : dict, default=None
        Fueltype lookup
    fueltypes_nr : int, default=None
        Number of modelled fueltypes
    """
    def __init__(
            self,
            lookup_enduses=None,
            lookup_sector_enduses=None,
            base_yr=None,
            weather_by=None,
            simulation_end_yr=None,
            curr_yr=None,
            sim_yrs=None,
            paths=None,
            enduses=None,
            sectors=None,
            reg_nrs=None
        ):
        """Constructor
        """
        self.lookup_enduses = lookup_enduses
        self.lookup_sector_enduses = lookup_sector_enduses

        self.submodels_names = lookup_tables.basic_lookups()['submodels_names']
        self.nr_of_submodels = len(self.submodels_names)
        self.fueltypes = lookup_tables.basic_lookups()['fueltypes']
        self.fueltypes_nr = lookup_tables.basic_lookups()['fueltypes_nr']

        self.base_yr = base_yr
        self.weather_by = weather_by
        self.reg_nrs = reg_nrs
        self.simulation_end_yr = simulation_end_yr
        self.curr_yr = curr_yr
        self.sim_yrs = sim_yrs

        # ============================================================
        # Spatially modelled variables
        #
        # If spatial explicit diffusion is modelled, all parameters
        # or technologies having a spatial explicit diffusion need
        # to be defined.
        # ============================================================
        self.spatial_explicit_diffusion = 0 #0: False, 1: True

        # Define all variables which are affected by regional diffusion
        self.spatially_modelled_vars = [] # ['smart_meter_p']

        # Define technologies which are affected by spatial explicit diffusion
        self.techs_affected_spatial_f = ['heat_pumps_electricity']

        # Max penetration speed
        self.speed_con_max = 1 #1.5 # 1: uniform distribution >1: regional differences

        # ============================================================
        # Model calibration factors
        # ============================================================
        #
        #   These calibration factors are used to match the modelled
        #   electrictiy demand better with the validation data.
        #
        #   Weekend effects are used to distribut energy demands
        #   between working and weekend days. With help of these
        #   factors, the demand on weekends and holidays can be
        #   be lowered compared to working days.
        #   This factor can be applied either directly to an enduse
        #   or to the hdd or cdd calculations (to correct cooling
        #   or heating demand)
        #
        #       f_ss_cooling_weekend : float
        #           Weekend effect for cooling enduses
        #       f_ss_weekend : float
        #           WWeekend effect for service submodel enduses
        #       f_is_weekend : float
        #           Weekend effect for industry submodel enduses
        #       f_mixed_floorarea : float
        #           Share of floor_area which is assigned to either
        #           residential or non_residential floor area
        # ------------------------------------------------------------
        self.f_ss_cooling_weekend = 0.45                # Temporal calibration factor
        self.f_ss_weekend = 0.8                         # Temporal calibration factor
        self.f_is_weekend = 0.45                        # Temporal calibration factor

        # ============================================================
        #   Modelled day related factors
        # ============================================================
        #   model_yeardays_date : dict
        #     Contains for the base year for each days
        #     the information wheter this is a working or holiday
        # ------------------------------------------------------------
        self.model_yeardays = list(range(365))

        # Calculate dates
        self.model_yeardays_date = []
        for yearday in self.model_yeardays:
            self.model_yeardays_date.append(
                date_prop.yearday_to_date(base_yr, yearday))

        # ============================================================
        #   Dwelling stock related assumptions
        # ============================================================
        #
        #   Assumptions to generate a virtual dwelling stock
        #
        #       assump_diff_floorarea_pp : float
        #           Change in floor area per person (%, 1=100%)
        #       assump_diff_floorarea_pp_yr_until_changed : int
        #           Year until this change in floor area happens
        #       dwtype_distr_by : dict
        #           Housing Stock Distribution by Type
        #               Source: UK Housing Energy Fact File, Table 4c
        #       dwtype_distr_fy : dict
        #           welling type distribution end year
        #               Source: UK Housing Energy Fact File, Table 4c
        #       dwtype_floorarea_by : dict
        #           Floor area per dwelling type (Annex Table 3.1)
        #               Source: UK Housing Energy Fact File, Table 4c
        #       dwtype_floorarea_fy : dict
        #           Floor area per dwelling type
        #               Source: UK Housing Energy Fact File, Table 4c
        #       dwtype_age_distr : dict
        #           Floor area per dwelling type
        #               Source: Housing Energy Fact Sheet)
        #       yr_until_changed : int
        #           Year until change is realised
        #
        # https://www.gov.uk/government/statistics/english-housing-survey-2014-to-2015-housing-stock-report
        # ------------------------------------------------------------
        yr_until_changed_all_things = 2050

        self.dwtype_distr_by = {
            'semi_detached': 0.26,
            'terraced': 0.283,
            'flat': 0.203,
            'detached': 0.166,
            'bungalow': 0.088}

        self.dwtype_distr_fy = {
            'yr_until_changed': yr_until_changed_all_things,

            'semi_detached': 0.26,
            'terraced': 0.283,
            'flat': 0.203,
            'detached': 0.166,
            'bungalow': 0.088}

        self.dwtype_floorarea_by = {
            'semi_detached': 96,
            'terraced': 82.5,
            'flat': 61,
            'detached': 147,
            'bungalow': 77}

        self.dwtype_floorarea_fy = {
            'yr_until_changed': yr_until_changed_all_things,

            'semi_detached': 96,
            'terraced': 82.5,
            'flat': 61,
            'detached': 147,
            'bungalow': 77}

        # (Average builing age within age class, fraction)
        # The newest category of 2015 is added to implement change in refurbishing rate
        # For the base year, this is set to zero (if e.g. with future scenario set to 5%, then
        # proportionally to base year distribution number of houses are refurbished)
        self.dwtype_age_distr = {
            2015: {
                '1918' :0.21,
                '1941': 0.36,
                '1977.5': 0.3,
                '1996.5': 0.08,
                '2002': 0.05}}

        # ============================================================
        #  Scenario drivers
        # ============================================================
        #
        #   For every enduse the relevant factors which affect enduse
        #   consumption can be added in a list.
        #
        #   Note:   If e.g. floorarea and population are added, the
        #           effects will be overestimates (i.e. no multi-
        #           collinearity are considered).
        #
        #   scenario_drivers : dict
        #     Scenario drivers per enduse
        # ------------------------------------------------------------
        self.scenario_drivers = {

            # --Residential
            'rs_space_heating': ['floorarea', 'hlc'], # Do not use HDD or pop because otherweise double count
            'rs_water_heating': ['population'],
            'rs_lighting': ['population', 'floorarea'],
            'rs_cooking': ['population'],
            'rs_cold': ['population'],
            'rs_wet': ['population'],
            'rs_consumer_electronics': ['population', 'gva'],
            'rs_home_computing': ['population'],

            # --Service
            'ss_space_heating': ['floorarea'],
            'ss_water_heating': ['population'],
            'ss_lighting': ['floorarea'],
            'ss_catering': ['population'],
            'ss_ICT_equipment': ['population'],
            'ss_cooling_humidification': ['floorarea', 'population'],
            'ss_fans': ['floorarea', 'population'],
            'ss_small_power': ['population'],
            'ss_cooled_storage': ['population'],
            'ss_other_gas': ['population'],
            'ss_other_electricity': ['population'],

            # Industry
            'is_high_temp_process': ['gva'],
            'is_low_temp_process': ['gva'],
            'is_drying_separation': ['gva'],
            'is_motors': ['gva'],
            'is_compressed_air': ['gva'],
            'is_lighting': ['gva'],
            'is_space_heating': ['gva'],
            'is_other': ['gva'],
            'is_refrigeration': ['gva']}

        # ============================================================
        #   Cooling related assumptions
        # ============================================================
        #   assump_cooling_floorarea : int
        #       The percentage of cooled floor space in the base year
        #
        #   Literature
        #   ----------
        #   Abela, A. et al. (2016). Study on Energy Use by Air
        #   Conditioning. Bre, (June), 31. Retrieved from
        #   https://www.bre.co.uk/filelibrary/pdf/projects/aircon-energy-use/StudyOnEnergyUseByAirConditioningFinalReport.pdf
        # ------------------------------------------------------------

        # See Abela et al. (2016) & Carbon Trust. (2012). Air conditioning. Maximising comfort, minimising energy consumption
        self.cooled_ss_floorarea_by = 0.35

        # ============================================================
        # Smart meter related base year assumptions
        # ============================================================
        #   smart_meter_p_by : int
        #       The percentage of households with smart meters in by
        # ------------------------------------------------------------
        self.smart_meter_assump = {}

        # Currently in 2017 8.6 mio smart meter installed of 27.2 mio households --> 31.6%
        # https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/671930/Smart_Meters_2017_update.pdf)
        # In 2015, 5.8 % percent of all househods had one: https://assets.publishing.service.gov.uk/government/uploads/system/uploads/attachment_data/file/533060/2016_Q1_Smart_Meters_Report.pdf 
        self.smart_meter_assump['smart_meter_p_by'] = 0.05

        # Long term smart meter induced general savings, purley as
        # a result of having a smart meter (e.g. 0.03 --> 3% savings)
        # DECC 2015: Smart Metering Early Learning Project: Synthesis report
        # https://www.gov.uk/government/publications/smart-metering-early-learning-project-and-small-scale-behaviour-trials
        # Reasonable assumption is between 0.03 and 0.01 (DECC 2015)
        self.smart_meter_assump['savings_smart_meter'] = {

            # Residential
            'rs_cold': 0.03,
            'rs_cooking': 0.03,
            'rs_lighting': 0.03,
            'rs_wet': 0.03,
            'rs_consumer_electronics': 0.03,
            'rs_home_computing': 0.03,
            'rs_space_heating': 0.03,
            'rs_water_heating': 0.03,

            # Service
            'ss_space_heating': 0.03,
            'ss_water_heating': 0.03,
            'ss_cooling_humidification': 0.03,
            'ss_fans': 0.03,
            'ss_lighting': 0.03,
            'ss_catering': 0.03,
            'ss_small_power': 0.03,
            'ss_ICT_equipment': 0.03,
            'ss_cooled_storage': 0.03,
            'ss_other_gas': 0.03,
            'ss_other_electricity': 0.03,

            # Industry submodule
            'is_high_temp_process': 0,
            'is_low_temp_process': 0,
            'is_drying_separation': 0,
            'is_motors': 0,
            'is_compressed_air': 0,
            'is_lighting': 0,
            'is_space_heating': 0,
            'is_other': 0,
            'is_refrigeration': 0}

        # ============================================================
        # Base temperature assumptions
        # ============================================================
        #
        #   Parameters related to smart metering
        #
        #   rs_t_heating : int
        #       Residential submodel base temp of heating of base year
        #   rs_t_cooling_by : int
        #       Residential submodel base temp of cooling of base year
        #   ...
        #
        #   Note
        #   ----
        #   Because demand for cooling cannot directly be linked to
        #   calculated cdd, the paramters 'ss_t_base_cooling' is used
        #   as a calibration factor. By artifiallcy lowering this
        #   parameter, the energy demand assignement over the days
        #   in a year is improved.
        # ------------------------------------------------------------
        t_bases = {
            'rs_t_heating': 15.5,
            'ss_t_heating': 15.5,
            'ss_t_cooling': 5,
            'is_t_heating': 15.5}

        self.t_bases = DummyClass(t_bases)

        # ============================================================
        # Enduses lists affed by hdd/cdd
        # ============================================================
        #
        #   These lists show for which enduses temperature related
        #   calculations are performed.
        #
        #   enduse_space_heating : list
        #       All enduses for which hdd are used for yd calculations
        #   ss_enduse_space_cooling : list
        #       All service submodel enduses for which cdd are used for
        #       yd calculations
        # ------------------------------------------------------------
        self.enduse_space_heating = [
            'rs_space_heating',
            'ss_space_heating',
            'is_space_heating']

        self.ss_enduse_space_cooling = [
            'ss_cooling_humidification']

        # ============================================================
        # Industry related
        #
        # High temperature processing (high_temp_ process) dominates
        # energy consumption in the iron and steel
        #
        # ---- Steel production - Enduse: is_high_temp_process, Sector: basic_metals
        # With industry service switch, the future shares of 'is_temp_high_process'
        # in sector 'basic_metals' can be set for 'basic_oxygen_furnace',
        # 'electric_arc_furnace', and 'SNG_furnace' can be specified
        #
        # ---- Cement production - Enduse: is_high_temp_process, Sector: non_metallic_mineral_products
        # Dry kilns, semidry kilns can be set
        # ============================================================

        # Share of cold rolling in steel manufacturing
        self.p_cold_rolling_steel_by = 0.2   # Estimated based on https://aceroplatea.es/docs/EuropeanSteelFigures_2015.pdf
        self.eff_cold_rolling_process = 1.8  # 80% more efficient than hot rolling Fruehan et al. (2002)
        self.eff_hot_rolling_process = 1.0   # 100% assumed efficiency

        # ============================================================
        # Assumption related to heat pump technologies
        # ============================================================
        #
        #   Assumptions related to technologies
        #
        #   gshp_fraction : list
        #       Fraction of installed gshp_fraction heat pumps in base year
        #       ASHP = 1 - gshp_fraction
        # ------------------------------------------------------------
        self.gshp_fraction = 0.1

        # Load defined technologies
        self.technologies, self.tech_list = read_data.read_technologies(paths['path_technologies'])

        self.installed_heat_pump_by = tech_related.generate_ashp_gshp_split(
            self.gshp_fraction)

        # Add heat pumps to technologies
        self.technologies, self.tech_list['heating_non_const'], self.heat_pumps = tech_related.generate_heat_pump_from_split(
            self.technologies,
            self.installed_heat_pump_by,
            self.fueltypes)

        # ============================================================
        # Fuel Stock Definition
        # Provide for every fueltype of an enduse the share of fuel
        # which is used by technologies in the base year
        # ============================================================$
        fuel_tech_p_by = fuel_shares.assign_by_fuel_tech_p(
            enduses,
            sectors,
            self.fueltypes,
            self.fueltypes_nr)

        # ========================================
        # Get technologies of an enduse and sector
        # ========================================
        self.specified_tech_enduse_by = helpers.get_def_techs(
            fuel_tech_p_by)

        _specified_tech_enduse_by = helpers.add_undef_techs(
            self.heat_pumps,
            self.specified_tech_enduse_by,
            self.enduse_space_heating)
        self.specified_tech_enduse_by = _specified_tech_enduse_by

        # ========================================
        # General other info
        # ========================================
        self.seasons = date_prop.get_season(year_to_model=base_yr)
        self.model_yeardays_daytype, self.yeardays_month, self.yeardays_month_days = date_prop.get_yeardays_daytype(
            year_to_model=base_yr)

        # ========================================
        # Helper functions
        # ========================================
        self.fuel_tech_p_by, self.specified_tech_enduse_by, self.technologies = tech_related.insert_placholder_techs(
            self.technologies,
            fuel_tech_p_by,
            self.specified_tech_enduse_by)

        # ========================================
        # Calculations with assumptions
        # ========================================
        self.cdd_weekend_cfactors = hdd_cdd.calc_weekend_corr_f(
            self.model_yeardays_daytype,
            self.f_ss_cooling_weekend)

        self.ss_weekend_f = hdd_cdd.calc_weekend_corr_f(
            self.model_yeardays_daytype,
            self.f_ss_weekend)

        self.is_weekend_f = hdd_cdd.calc_weekend_corr_f(
            self.model_yeardays_daytype,
            self.f_is_weekend)

        # ========================================
        # Testing
        # ========================================
        testing_functions.testing_fuel_tech_shares(
            self.fuel_tech_p_by)

        testing_functions.testing_tech_defined(
            self.technologies, self.specified_tech_enduse_by)

    def update(self, name, value):
        """Update assumptions

        Arguments
        ---------
        name : str
            name of attribute
        value : any
            Type of value
        """
        setattr(self, name, value)

def update_technology_assumption(
        technologies,
        narrative_f_eff_achieved,
        narrative_gshp_fraction,
        crit_narrative_input=True
    ):
    """Updates technology related properties based on
    scenario assumptions. Calculate average efficiency of
    heat pumps depending on mix of GSHP and ASHP,
    set the efficiency achieval factor of all factor according
    to strategy assumptions

    Parameters
    ----------
    technologies : dict
        Technologies
    f_eff_achieved : float
        Factor achieved
    gshp_fraction : float
        Mix of GSHP and GSHP
    crit_narrative_input : bool
        Criteria wheter inputs are single values or a narrative

    Note
    ----
    This needs to be run everytime an assumption is changed
    """
    if crit_narrative_input:
        # Read from narrative the value
        f_eff_achieved = narrative_related.read_from_narrative(narrative_f_eff_achieved) 
        gshp_fraction = narrative_related.read_from_narrative(narrative_gshp_fraction)
    else:
        f_eff_achieved = narrative_f_eff_achieved
        gshp_fraction = narrative_gshp_fraction

    # Assign same achieved efficiency factor for all technologies
    technologies = helpers.set_same_eff_all_tech(
        technologies,
        f_eff_achieved)

    # Calculate average eff of hp depending on fraction of GSHP to ASHP
    installed_heat_pump_ey = tech_related.generate_ashp_gshp_split(
        gshp_fraction)

    technologies = tech_related.calc_av_heat_pump_eff_ey(
        technologies, installed_heat_pump_ey)

    return technologies

class DummyClass(object):
    """Assumptions
    """
    def __init__(
            self,
            variables
        ):
        for var, value in variables.items():
            setattr(self, var, value)



"""testing
"""
from energy_demand.scripts import init_scripts

def test_sum_across_sectors_all_regs():
    """testing"""

    fuel_disagg_reg = {
        'regA': {
            'enduse': {'sectorA': 100, 'sectorB': 100}}}

    out = init_scripts.sum_across_sectors_all_regs(fuel_disagg_reg)

    assert out['regA'] == {'enduse': 200}



"""This scripts reads the national electricity data for the base year"""
import os
import csv
import logging
import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab

from energy_demand.basic import date_prop
from energy_demand.basic import conversions
from energy_demand.plotting import basic_plot_functions
from energy_demand.basic import basic_functions
from energy_demand.plotting import plotting_styles

def read_raw_elec_2015(path_to_csv, year=2015):
    """Read in national electricity values provided
    in MW and convert to GWh

    Arguments
    ---------
    path_to_csv : str
        Path to csv file
    year : int
        Year of data

    Returns
    -------
    elec_data_indo : array
        Hourly INDO electricity in GWh (INDO - National Demand)
    elec_data_itsdo : array
        Hourly ITSDO electricity in GWh (Transmission System Demand)

    Note
    -----
    Half hourly measurements are aggregated to hourly values

    Necessary data preparation: On 29 March and 25 Octobre
    there are 46 and 48 values because of the changing of the clocks
    The 25 Octobre value is omitted, the 29 March hour interpolated
    in the csv file

    Source
    ------
    http://www2.nationalgrid.com/uk/Industry-information/electricity-transmission-operational-data/
    For more information on INDO and ISTDO see DemandData Field Descriptions file:
    http://www2.nationalgrid.com/WorkArea/DownloadAsset.aspx?id=8589934632

    National Demand is calculated as a sum
    of generation based on National Grid
    operational generation metering
    """
    elec_data_indo = np.zeros((365, 24), dtype="float")
    elec_data_itsdo = np.zeros((365, 24), dtype="float")

    with open(path_to_csv, 'r') as csvfile:
        read_lines = csv.reader(csvfile, delimiter=',')
        _headings = next(read_lines)

        hour = 0
        counter_half_hour = 0

        for line in read_lines:
            month = basic_functions.get_month_from_string(line[0].split("-")[1])
            day = int(line[0].split("-")[0])

            # Get yearday
            yearday = date_prop.date_to_yearday(year, month, day)

            if counter_half_hour == 1:
                counter_half_hour = 0

                # Sum value of first and second half hour
                hour_elec_demand_INDO = half_hour_demand_indo + float(line[2])
                hour_elec_demand_ITSDO = half_hour_demand_itsdo + float(line[4])

                # Convert MW to GWH (input is MW aggregated for two half
                # hourly measurements, therfore divide by 0.5)
                elec_data_indo[yearday][hour] = conversions.mw_to_gwh(hour_elec_demand_INDO, 0.5)
                elec_data_itsdo[yearday][hour] = conversions.mw_to_gwh(hour_elec_demand_ITSDO, 0.5)

                hour += 1
            else:
                counter_half_hour += 1

                half_hour_demand_indo = float(line[2]) # INDO - National Demand
                half_hour_demand_itsdo = float(line[4]) # Transmission System Demand

            if hour == 24:
                hour = 0

    return elec_data_indo, elec_data_itsdo

def compare_results(
        name_fig,
        path_result,
        y_factored_indo,
        y_calculated_array,
        title_left,
        days_to_plot,
        plot_crit=False
    ):
    """Compare national electrictiy demand data with model results

    Note
    ----
    RMSE fit criteria : Lower values of RMSE indicate better fit
    https://stackoverflow.com/questions/17197492/root-mean-square-error-in-python

    https://matplotlib.org/examples/lines_bars_and_markers/marker_fillstyle_reference.html
    """
    logging.debug("...compare elec results")
    nr_of_h_to_plot = len(days_to_plot) * 24

    x_data = range(nr_of_h_to_plot)

    y_real_indo_factored = []
    y_calculated_list = []
    y_diff_p = []
    y_diff_abs = []

    for day in days_to_plot:
        for hour in range(24):
            y_calculated_list.append(y_calculated_array[day][hour])
            y_real_indo_factored.append(y_factored_indo[day][hour])

            # Calculate absolute differences
            abs_diff = abs(y_factored_indo[day][hour] - y_calculated_array[day][hour])
            y_diff_abs.append(abs_diff)

            # Calculate difference in percent
            if abs_diff == 0:
                p_diff = 0
            else:
                p_diff = (100 / y_factored_indo[day][hour]) * y_calculated_array[day][hour] - 100

            y_diff_p.append(p_diff)

    # -------------
    # RMSE
    # -------------
    rmse_val_corrected = basic_functions.rmse(
        np.array(y_real_indo_factored),
        np.array(y_calculated_list))

    # ----------
    # Standard deviation
    # ----------
    standard_dev_real_modelled = np.std(y_diff_p)       # Differences in %
    standard_dev_real_modelled_abs = np.std(y_diff_abs) # Absolute differences 

    logging.info(
        "Standard deviation given as percentage: " + str(standard_dev_real_modelled))
    logging.info(
        "Standard deviation given as GW:         " + str(standard_dev_real_modelled_abs))

    # ---------
    # R squared
    # ---------
    slope, intercept, r_value, p_value, std_err = stats.linregress(
        y_real_indo_factored,
        y_calculated_list)

    # ----------
    # Plot residuals
    # ----------
    #try:
    #    plot_residual_histogram(
    #        y_diff_p, path_result, "residuals_{}".format(name_fig))
    #except:
    #    pass

    # ----------
    # Plot figure
    # ----------
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(22, 8))

    # smooth line
    x_data_smoothed, y_real_indo_factored_smoothed = basic_plot_functions.smooth_data(
        x_data, y_real_indo_factored, num=40000)

    # plot points
    plt.plot(
        x_data_smoothed,
        y_real_indo_factored_smoothed,
        label='indo_factored',
        linestyle='-',
        linewidth=0.5,
        fillstyle='full',
        color='black')

    # smooth line
    x_data_smoothed, y_calculated_list_smoothed = basic_plot_functions.smooth_data(
        x_data, y_calculated_list, num=40000)

    plt.plot(
        x_data_smoothed,
        y_calculated_list_smoothed,
        label='model',
        linestyle='--',
        linewidth=0.5,
        fillstyle='full',
        color='blue')

    plt.xlim([0, 8760])
    plt.margins(x=0)
    plt.axis('tight')

    # --------------------------------------
    # Label x axis in dates
    # --------------------------------------
    major_ticks_days, major_ticks_labels = get_date_strings(
        days_to_plot,
        daystep=1)

    plt.xticks(major_ticks_days, major_ticks_labels)

    # ----------
    # Labelling
    # ----------
    font_additional_info = plotting_styles.font_info(size=4)

    plt.title(
        'RMSE: {} Std_dev_% {} (+-{} GW) R_2: {}'.format(
            round(rmse_val_corrected, 3),
            round(standard_dev_real_modelled, 3),
            round(standard_dev_real_modelled_abs, 3),
            round(r_value, 3)),
        fontdict=font_additional_info,
        loc='right')

    plt.title(title_left, loc='left')

    plt.xlabel("hour", fontsize=10)
    plt.ylabel("uk elec use [GW] for {}".format(title_left), fontsize=10)

    plt.legend(frameon=False)

    plt.savefig(os.path.join(path_result, name_fig))

    if plot_crit:
        plt.show()
    plt.close()

def compare_peak(
        name_fig,
        path_result,
        real_elec_2015_peak,
        modelled_peak_dh,
        peak_day
    ):
    """Compare peak electricity day with calculated peak energy demand

    Arguments
    ---------
    name_fig : str
        Name of figure
    local_paths : dict
        Paths
    real_elec_2015_peak : array
        Real data of peak day
    modelled_peak_dh : array
        Modelled peak day
    """
    logging.debug("...compare elec peak results")

    real_elec_peak = np.copy(real_elec_2015_peak)
    # -------------------------------
    # Compare values
    # -------------------------------
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(8, 8))

    # smooth line
    x_smoothed, y_modelled_peak_dh_smoothed = basic_plot_functions.smooth_data(range(24), modelled_peak_dh, num=500)

    plt.plot(
        x_smoothed,
        y_modelled_peak_dh_smoothed,
        color='blue',
        linestyle='--',
        linewidth=0.5,
        label='model')

    x_smoothed, real_elec_peak_smoothed = basic_plot_functions.smooth_data(
        range(24), real_elec_peak, num=500)

    plt.plot(
        x_smoothed,
        real_elec_peak_smoothed,
        color='black',
        linestyle='-',
        linewidth=0.5,
        label='validation')

    #raise Exception
    # Calculate hourly differences in %
    diff_p_h = np.round((100 / real_elec_peak) * modelled_peak_dh, 1)

    # Calculate maximum difference
    max_h_real = np.max(real_elec_peak)
    max_h_modelled = np.max(modelled_peak_dh)

    max_h_diff = round((100 / max_h_real) * max_h_modelled, 2)
    max_h_diff_gwh = round((abs(100 - max_h_diff)/100) * max_h_real, 2)

    # Y-axis ticks
    plt.xlim(0, 23)
    plt.yticks(range(0, 60, 10))

    # because position 0 in list is 01:00, the labelling starts with 1
    plt.xticks([0, 5, 11, 17, 23], [1, 6, 12, 18, 24]) #ticks, labels

    plt.legend(frameon=False)

    # Labelling
    date_yearday = date_prop.yearday_to_date(2015, peak_day)
    plt.title("peak comparison {}".format(date_yearday))

    plt.xlabel("h (max {} ({} GWH)".format(max_h_diff, max_h_diff_gwh))
    plt.ylabel("uk electrictiy use [GW]")

    plt.text(
        5, #position
        5, #position
        diff_p_h,
        fontdict={
            'family': 'arial',
            'color': 'black',
            'weight': 'normal',
            'size': 4})

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    # Save fig
    plt.savefig(os.path.join(path_result, name_fig))
    plt.close()

def compare_results_hour_boxplots(
        name_fig,
        path_result,
        data_real,
        data_calculated
    ):
    """Calculate differences for everyhour and plot according
    to hour for the full year
    """
    data_h_full_year = {}
    for i in range(24):
        data_h_full_year[i] = []

    for yearday_python in range(365):
        for hour in range(24):

            # Differenc in % of real value
            diff_percent = (100 / data_real[yearday_python][hour]) * data_calculated[yearday_python][hour]

            # Add differene to list of specific hour
            data_h_full_year[hour].append(diff_percent)

    fig = plt.figure()
    ax = fig.add_subplot(111)

    # Add a horizontal grid to the plot, but make it very light in color so we
    # can use it for reading data values but not be distracting
    ax.yaxis.grid(True, linestyle='-', which='major', color='lightgrey', alpha=0.5)
    ax.axhline(y=100, xmin=0, xmax=3, c="red", linewidth=1, zorder=0)

    diff_values = []
    for hour in range(24):
        diff_values.append(np.asarray(data_h_full_year[hour]))

    ax.boxplot(diff_values)

    plt.xticks(range(1, 25), range(24))

    plt.xlabel("hour")
    plt.ylabel("Modelled electricity difference (real-modelled) [%]")

    plt.savefig(os.path.join(path_result, name_fig))
    plt.close()

def plot_residual_histogram(values, path_result, name_fig):
    """Plot residuals as histogram and test for normal distribution

    https://matplotlib.org/1.2.1/examples/pylab_examples/histogram_demo.html
    https://stackoverflow.com/questions/22179119/normality-test-of-a-distribution-in-python
    """
    plt.figure(facecolor="white")

    # Sort according to size
    values.sort()

    # the histogram of the data
    n, bins, patches = plt.hist(
        values,
        50,
        normed=1,
        facecolor='grey')

    # ------------------
    # Plot normal distribution
    # ------------------
    mu = np.average(values)
    sigma = np.std(values)
    plt.plot(
        bins,
        mlab.normpdf(bins, mu, sigma),
        color='r')

    # -----------
    # Test for normal distribution
    # https://stackoverflow.com/questions/12838993/scipy-normaltest-how-is-it-used
    # http://www.socscistatistics.com/pvalues/chidistribution.aspx
    # http://stattrek.com/chi-square-test/goodness-of-fit.aspx?Tutorial=AP
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.normaltest.html
    # -----------
    chi_squared, p_value = stats.normaltest(values)

    # ---------------
    # plot histogram
    # ---------------
    font_additional_info = plotting_styles.font_info()

    plt.xlabel('Smarts')
    plt.ylabel('Probability')
    plt.title("Residual distribution (chi_squared: {}  p_value:  {}".format(
        round(chi_squared, 4),
        round(p_value, 4)),
              fontsize=10,
              fontdict=font_additional_info,
              loc='right')

    plt.savefig(os.path.join(path_result, name_fig))

def get_date_strings(days_to_plot, daystep):
    """Calculate date and position for range input of yeardays

    Arguments
    ---------
    days_to_plot : list
        List with yeardays to plot
    daystep : int
        Intervall of days to assign label

    Return
    -------
    ticks_position : list
        Hourly ticks position
    major_ticks_labels : list
        Ticks labels
    """
    ticks_position, ticks_labels = [], []

    cnt = 0
    cnt_daystep = 1

    for day in days_to_plot:
        str_date = str(date_prop.yearday_to_date(2015, day))
        str_date_short = str_date[5:]

        # Because 0 posit in hour list is 01:00, substract minus one hour
        yearhour = (cnt * 24) - 1

        if daystep == cnt_daystep:
            ticks_labels.append(str_date_short)
            ticks_position.append(yearhour)

            cnt_daystep = 1
            cnt += 1
        else:
            cnt_daystep += 1
            cnt += 1

    return ticks_position, ticks_labels



"""Testing
"""
from energy_demand.scripts import s_fuel_to_service
from energy_demand.technologies import technological_stock
from energy_demand.read_write import read_data
import numpy as np

def init_nested_dict_brackets():
    """Testing"""
    result = s_fuel_to_service.init_nested_dict_brackets(
        first_level_keys=["A", "B"],
        second_level_keys=[1, 2])

    expected = {"A": {1: {}, 2: {}}, "B": {1: {}, 2: {}}}

    assert result == expected
    assert result[0] == expected[0]

def test_init_nested_dict_zero():
    """Testing"""
    result = s_fuel_to_service.init_nested_dict_zero(
        sector=None,
        first_level_keys=["A", "B"],
        second_level_keys=[1, 2])

    expected = {"A": {None: {1: 0, 2: 0}}, "B": {None: {1: 0, 2: 0}}}

    assert result == expected

def test_sum_2_level_dict():
    """Testing"""
    two_level_dict = {"A": {1: 30, 2: 2}, "B": {1: 1, 2: 0}}

    result = s_fuel_to_service.sum_2_level_dict(two_level_dict)

    assert result == 33

def test_get_s_fueltype_tech():
    """
    """
    tech_list = { 
        'heating_non_const': ['heat_p'],
        'heating_const': ['boilerA'],
        'storage_heating_electricity': ['boilerC'],
        'secondary_heating_electricity': []}

    fueltypes = {
        'gas': 0} 

    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0)}

    fuel_p_tech_by = {'heating': {'sectorA': {0: {'boilerA': 1.0}}}}

    s_tech_by_p, s_fueltype_by_p = s_fuel_to_service.get_s_fueltype_tech(
       enduses=['heating'],
       fuel_p_tech_by=fuel_p_tech_by,
       fuels={'heating': {'sectorA': np.array([100])}},
       technologies=technologies,
       sector='sectorA')

    assert s_tech_by_p['heating']['boilerA'] == 1.0
    assert s_fueltype_by_p['heating']['sectorA'][0] == 1.0

    # -------------------------------------
    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0),
        'boilerB': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0)}

    fuel_p_tech_by = {'heating': {'sectorA': {0: {'boilerA': 0.5, 'boilerB': 0.5}}}}

    s_tech_by_p, s_fueltype_by_p = s_fuel_to_service.get_s_fueltype_tech(
       enduses=['heating'],
       fuel_p_tech_by=fuel_p_tech_by,
       fuels={'heating': {'sectorA': np.array([100])}}, # 0 == gas
       technologies=technologies,
       sector='sectorA')

    assert s_tech_by_p['heating']['boilerA'] == 0.5
    assert s_tech_by_p['heating']['boilerB'] == 0.5
    assert s_fueltype_by_p['heating']['sectorA'][0] == 1.0

    # -------------------------------------

    tech_list = { 
        'heating_non_const': ['heat_p'],
        'heating_const': ['boilerA', 'boilerB'],
        'storage_heating_electricity': [],
        'secondary_heating_electricity': []}
    
    fueltypes = {
        'gas': 0,
        'electricity': 1} 

    technologies = {
        'boilerA': read_data.TechnologyData(
            fueltype='gas',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0),
        'boilerB': read_data.TechnologyData(
            fueltype='electricity',
            eff_by=0.5,
            eff_ey=0.5,
            year_eff_ey=2015,
            eff_achieved=1.0,
            diff_method='linear',
            market_entry=1990,
            tech_max_share=1.0)}

    fuel_p_tech_by = {'heating': {'sectorA': {0: {'boilerA': 1.0}, 1: {'boilerB': 1.0}}}}

    s_tech_by_p, s_fueltype_by_p = s_fuel_to_service.get_s_fueltype_tech(
       enduses=['heating'],
       fuel_p_tech_by=fuel_p_tech_by,
       fuels={'heating': {'sectorA': np.array([100, 300])}}, # tripple elec than gas
       technologies=technologies,
       sector='sectorA')

    assert s_tech_by_p['heating']['boilerA'] == 0.25
    assert s_tech_by_p['heating']['boilerB'] == 0.75
    assert s_fueltype_by_p['heating']['sectorA'][0] == 0.25
    assert s_fueltype_by_p['heating']['sectorA'][1] == 0.75
